# Copyright (C) 2004-2011 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test container routines.
"""

import unittest

import linkcheck.containers


class TestLFUCache(unittest.TestCase):
    """Test LFU cache implementation."""

    def setUp(self):
        """Set up self.d as empty LFU cache with default size of 1000."""
        self.size = 1000
        self.d = linkcheck.containers.LFUCache(self.size)

    def test_num_uses(self):
        self.assertTrue(not self.d)
        self.d["a"] = 1
        self.assertTrue("a" in self.d)
        self.assertEqual(self.d.uses("a"), 0)
        dummy = self.d["a"]
        self.assertEqual(self.d.uses("a"), 1)
        del dummy

    def test_values(self):
        self.assertTrue(not self.d)
        self.d["a"] = 1
        self.d["b"] = 2
        self.assertEqual({1, 2}, set(self.d.values()))
        self.assertEqual({1, 2}, set(self.d.itervalues()))

    def test_popitem(self):
        self.assertTrue(not self.d)
        self.d["a"] = 42
        self.assertEqual(self.d.popitem(), ("a", 42))
        self.assertTrue(not self.d)
        self.assertRaises(KeyError, self.d.popitem)

    def test_shrink(self):
        self.assertTrue(not self.d)
        for i in range(self.size):
            self.d[i] = i
        self.d[1001] = 1001
        self.assertTrue(950 <= len(self.d) <= self.size)



# Copyright (C) 2020 Chris Mayo
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test LocationInfo plugin.

Requires GeoIP (libgeoip-dev to build) and geoip-database
"""

import unittest

from linkcheck.plugins import locationinfo
from tests import need_geoip


class TestLocationInfo(unittest.TestCase):
    @need_geoip
    def test_get_location(self):
        location = locationinfo.get_location("8.8.8.8")
        self.assertEqual(location, "United States")



# Copyright (C) 2014 Bastian Kleineidam
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import unittest
import sys
from . import run_checked


def run_with_options(options):
    """Run a command with given options."""
    run_checked([sys.executable, "-m", "linkcheck"] + options)


class TestLinkchecker(unittest.TestCase):
    """Test the linkchecker commandline client."""

    def test_linkchecker(self):
        # test some single options
        for option in (
                "-V",
                "--version",
                "-h",
                "--help",
                "--list-plugins",
                "-Dall",
                "-ocsv",
                "-Fhtml",
                ):
            run_with_options([option])
        # unknown option
        self.assertRaises(OSError, run_with_options, ["--imadoofus"])
        # non-existent FILENAMEs
        self.assertRaises(OSError, run_with_options, ["--config", "no_such_file"])
        self.assertRaises(OSError, run_with_options, ["--cookiefile", "no_such_file"])



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Define http test support classes for LinkChecker tests.
"""

import html
from http.server import SimpleHTTPRequestHandler, HTTPServer
from http.client import HTTPConnection, HTTPSConnection
import ssl
import time
import threading
import urllib.parse
from io import BytesIO
from . import LinkCheckTest
from .. import get_file


class StoppableHttpRequestHandler(SimpleHTTPRequestHandler):
    """
    HTTP request handler with QUIT stopping the server.
    """

    def do_QUIT(self):
        """
        Send 200 OK response, and set server.stop to True.
        """
        self.send_response(200)
        self.end_headers()
        self.server.stop = True

    def log_message(self, format, *args):
        """
        Logging is disabled.
        """
        pass


# serve .xhtml files as application/xhtml+xml
StoppableHttpRequestHandler.extensions_map.update(
    {".xhtml": "application/xhtml+xml"}
)


class StoppableHttpServer(HTTPServer):
    """
    HTTP server that reacts to self.stop flag.
    """

    def serve_forever(self):
        """
        Handle one request at a time until stopped.
        """
        self.stop = False
        while not self.stop:
            self.handle_request()


class NoQueryHttpRequestHandler(StoppableHttpRequestHandler):
    """
    Handler ignoring the query part of requests and sending dummy directory
    listings.
    """

    def remove_path_query(self):
        """
        Remove everything after a question mark.
        """
        i = self.path.find("?")
        if i != -1:
            self.path = self.path[:i]

    def get_status(self):
        dummy, status = self.path.rsplit("/", 1)
        status = int(status)
        if status in self.responses:
            return status
        return 500

    def do_GET(self):
        """
        Removes query part of GET request.
        """
        self.remove_path_query()
        if "status/" in self.path:
            status = self.get_status()
            self.send_response(status)
            self.end_headers()
            if status >= 200 and status not in (204, 304):
                self.wfile.write(b"testcontent")
        else:
            super().do_GET()

    def do_HEAD(self):
        """
        Removes query part of HEAD request.
        """
        self.remove_path_query()
        if "status/" in self.path:
            self.send_response(self.get_status())
            self.end_headers()
        else:
            super().do_HEAD()

    def list_directory(self, path):
        """Helper to produce a directory listing (absent index.html).

        Return value is either a file object, or None (indicating an
        error).  In either case, the headers are sent, making the
        interface the same as for send_head().

        """
        f = BytesIO()
        f.write(b'<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">')
        f.write(b"<html>\n<title>Dummy directory listing</title>\n")
        f.write(b"<body>\n<h2>Dummy test directory listing</h2>\n")
        f.write(b"<hr>\n<ul>\n")
        list = ["example1.txt", "example2.html", "example3"]
        for name in list:
            displayname = linkname = name
            list_item = '<li><a href="{}">{}</a>\n'.format(
                urllib.parse.quote(linkname),
                html.escape(displayname),
            )
            f.write(list_item.encode())
        f.write(b"</ul>\n<hr>\n</body>\n</html>\n")
        length = f.tell()
        f.seek(0)
        self.send_response(200)
        encoding = "utf-8"
        self.send_header("Content-type", "text/html; charset=%s" % encoding)
        self.send_header("Content-Length", str(length))
        self.end_headers()
        return f


class HttpServerTest(LinkCheckTest):
    """
    Start/stop an HTTP server that can be used for testing.
    """

    def __init__(self, methodName="runTest"):
        """
        Init test class and store default http server port.
        """
        super().__init__(methodName=methodName)
        self.port = None
        self.handler = NoQueryHttpRequestHandler

    def setUp(self, https=False):
        """Start a new HTTP server in a new thread."""
        super().setUp()
        self.port = start_server(self.handler, https)
        assert self.port is not None

    def tearDown(self):
        """Send QUIT request to http server."""
        stop_server(self.port)

    def get_url(self, filename):
        """Get HTTP URL for filename."""
        return "http://localhost:%d/tests/checker/data/%s" % (self.port, filename)


class HttpsServerTest(HttpServerTest):
    """
    Start/stop an HTTPS server that can be used for testing.
    """

    def setUp(self):
        """Start a new HTTPS server in a new thread."""
        super().setUp(https=True)

    def tearDown(self):
        """Send QUIT request to http server."""
        stop_server(self.port, https=True)

    def get_url(self, filename):
        """Get HTTP URL for filename."""
        return "https://localhost:%d/tests/checker/data/%s" % (self.port, filename)


def start_server(handler, https=False):
    """Start an HTTP server thread and return its port number."""
    server_address = ("localhost", 0)
    handler.protocol_version = "HTTP/1.0"
    httpd = StoppableHttpServer(server_address, handler)
    if https:
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        context.load_cert_chain(
            get_file("https_cert.pem"), keyfile=get_file("https_key.pem"))
        httpd.socket = context.wrap_socket(
            httpd.socket,
            server_side=True,
        )
    port = httpd.server_port
    t = threading.Thread(None, httpd.serve_forever)
    t.start()
    # wait for server to start up
    while True:
        try:
            if https:
                conn = HTTPSConnection(
                    "localhost:%d" % port, context=ssl._create_unverified_context()
                )
            else:
                conn = HTTPConnection("localhost:%d" % port)
            conn.request("GET", "/")
            conn.getresponse()
            break
        except Exception:
            time.sleep(0.5)
    return port


def stop_server(port, https=False):
    """Stop an HTTP server thread."""
    if https:
        conn = HTTPSConnection(
            "localhost:%d" % port, context=ssl._create_unverified_context()
        )
    else:
        conn = HTTPConnection("localhost:%d" % port)
    conn.request("QUIT", "/")
    conn.getresponse()


def get_cookie(maxage=2000):
    data = (
        ("Comment", "justatest"),
        ("Max-Age", "%d" % maxage),
        ("Path", "/"),
        ("Version", "1"),
        ("Foo", "Bar"),
    )
    return "; ".join(f'{key}="{value}"' for key, value in data)


class CookieRedirectHttpRequestHandler(NoQueryHttpRequestHandler):
    """Handler redirecting certain requests, and setting cookies."""

    def end_headers(self):
        """Send cookie before ending headers."""
        self.send_header("Set-Cookie", get_cookie())
        self.send_header("Set-Cookie", get_cookie(maxage=0))
        super().end_headers()

    def redirect(self):
        """Redirect request."""
        path = self.path.replace("redirect", "newurl")
        self.send_response(302)
        self.send_header("Location", path)
        self.end_headers()

    def redirect_newhost(self):
        """Redirect request to a new host."""
        path = "http://www.example.com/"
        self.send_response(302)
        self.send_header("Location", path)
        self.end_headers()

    def redirect_newscheme(self):
        """Redirect request to a new scheme."""
        if "file" in self.path:
            path = "file:README.md"
        else:
            path = "ftp://example.com/"
        self.send_response(302)
        self.send_header("Location", path)
        self.end_headers()

    def do_GET(self):
        """Handle redirections for GET."""
        if "redirect_newscheme" in self.path:
            self.redirect_newscheme()
        elif "redirect_newhost" in self.path:
            self.redirect_newhost()
        elif "redirect" in self.path:
            self.redirect()
        else:
            super().do_GET()

    def do_HEAD(self):
        """Handle redirections for HEAD."""
        if "redirect_newscheme" in self.path:
            self.redirect_newscheme()
        elif "redirect_newhost" in self.path:
            self.redirect_newhost()
        elif "redirect" in self.path:
            self.redirect()
        else:
            super().do_HEAD()



#!/usr/bin/env python
# Copyright (C) 2012-2014 Bastian Kleineidam
"""Remove all lines after a given marker line.
"""
import fileinput
import sys


def main(args):
    """Remove lines after marker."""
    filename = args[0]
    marker = args[1]
    for line in fileinput.input(filename, inplace=1):
        print(line.rstrip())
        if line.startswith(marker):
            break


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright (C) 2005-2010 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test decorators.
"""

from io import StringIO
import time
import unittest

import linkcheck.decorators


class TestDecorators(unittest.TestCase):
    """
    Test decorators.
    """

    def test_timeit(self):
        @linkcheck.decorators.timed()
        def f():
            return 42

        self.assertEqual(f(), 42)

    def test_timeit2(self):
        log = StringIO()

        @linkcheck.decorators.timed(log=log, limit=0)
        def f():
            time.sleep(1)
            return 42

        self.assertEqual(f(), 42)
        self.assertRegex(log.getvalue(), r"f took 1\.0\d seconds")



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test file parsing.
"""
import os
from pathlib import Path
import sys
import zipfile

import pytest

from tests import need_network, need_word, need_pdflib
from . import LinkCheckTest, get_file


def unzip(filename, targetdir):
    """Unzip given zipfile into targetdir."""
    # There are likely problems with zipfile and non-Unicode filenames
    # https://github.com/python/cpython/issues/83042
    # https://github.com/python/cpython/issues/72267
    # https://github.com/python/cpython/issues/95463
    zf = zipfile.ZipFile(filename)
    for name in zf.namelist():
        if name.endswith("/"):
            os.mkdir(os.path.join(targetdir, name), 0o700)
        else:
            outfile = open(os.path.join(targetdir, name), "wb")
            try:
                outfile.write(zf.read(name))
            finally:
                outfile.close()


class TestFile(LinkCheckTest):
    """
    Test file:// link checking (and file content parsing).
    """

    def test_html(self):
        self.file_test("file.html")

    @need_network
    def test_html_url_quote(self):
        self.file_test("file_url_quote.html")

    def test_wml(self):
        self.file_test("file.wml")

    def test_text(self):
        self.file_test("file.txt")

    def test_asc(self):
        self.file_test("file.asc")

    def test_css(self):
        self.file_test("file.css")

    def test_php(self):
        self.file_test("file.php")

    def test_empty(self):
        self.file_test("empty.html")

    @need_word
    def test_word(self):
        confargs = dict(enabledplugins=["WordParser"])
        self.file_test("file.doc", confargs=confargs)

    @need_pdflib
    def test_pdf(self):
        confargs = dict(enabledplugins=["PdfParser"])
        self.file_test("file.pdf", confargs=confargs)

    def test_markdown(self):
        confargs = dict(enabledplugins=["MarkdownCheck"])
        if sys.version_info < (3, 12, 5):
            markdown_file = "file-octet-stream.markdown"
        else:
            markdown_file = "file.markdown"
        self.file_test(markdown_file, confargs=confargs)

    def test_urllist(self):
        self.file_test("urllist.txt")

    @pytest.mark.xfail(strict=True)
    def test_directory_listing(self):
        # unpack non-unicode filename which cannot be stored
        # in the SF subversion repository
        if os.name != "posix" or sys.platform != "linux":
            pytest.skip("Not running on POSIX or Linux")
        dirname = get_file("dir")
        if not os.path.isdir(dirname):
            unzip(dirname + ".zip", os.path.dirname(dirname))
        self.file_test("dir")

    def test_directory_listing_unicode(self):
        if os.name != "posix" or sys.platform != "linux":
            pytest.skip("Not running on POSIX or Linux")
        dirname = Path(get_file("udir"))
        dirname.mkdir(exist_ok=True)
        Path(dirname, "í»­¯¿.dat").touch()
        self.file_test("udir")

    def test_unicode_filename(self):
        # a unicode filename
        self.file_test("Мошкова.bin")

    def test_good_file(self):
        url = "file://%(curdir)s/%(datadir)s/file.txt" % self.get_attrs()
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "valid",
        ]
        self.direct(url, resultlines)

    def test_bad_file(self):
        if os.name == "nt":
            # Fails on NT platforms and I am too lazy to fix
            # Cause: url get quoted %7C which gets lowercased to
            # %7c and this fails.
            pytest.skip("Running on Windows")
        url = "file:/%(curdir)s/%(datadir)s/file.txt" % self.get_attrs()
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "error",
        ]
        self.direct(url, resultlines)

    def test_good_file_missing_dslash(self):
        # good file (missing double slash)
        attrs = self.get_attrs()
        url = "file:%(curdir)s/%(datadir)s/file.txt" % attrs
        resultlines = [
            "url %s" % url,
            "cache key file://%(curdir)s/%(datadir)s/file.txt" % attrs,
            "real url file://%(curdir)s/%(datadir)s/file.txt" % attrs,
            "valid",
        ]
        self.direct(url, resultlines)

    def test_good_dir(self):
        url = "file://%(curdir)s/%(datadir)s/" % self.get_attrs()
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
            "valid",
        ]
        self.direct(url, resultlines)

    def test_good_dir_space(self):
        url = "file://%(curdir)s/%(datadir)s/a b/" % self.get_attrs()
        nurl = self.norm(url)
        url2 = "file://%(curdir)s/%(datadir)s/a b/el.html" % self.get_attrs()
        nurl2 = self.norm(url2)
        url3 = "file://%(curdir)s/%(datadir)s/a b/t.txt" % self.get_attrs()
        nurl3 = self.norm(url3)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "valid",
            "url el.html",
            "cache key %s" % nurl2,
            "real url %s" % nurl2,
            "name el.html",
            "valid",
            "url t.txt",
            "cache key %s" % nurl3,
            "real url %s" % nurl3,
            "name t.txt",
            "valid",
        ]
        self.direct(url, resultlines, recursionlevel=2)



# Copyright (C) 2005-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
import signal
import subprocess
import os
import sys
import socket
import unittest
import pytest
from contextlib import contextmanager
from functools import lru_cache, wraps
from linkcheck import init_i18n, LinkCheckerInterrupt


class TestBase(unittest.TestCase):
    """
    Base class for tests.
    """

    def setUp(self):
        """Ensure the current locale setting is the default.
        Otherwise, warnings will get translated and will break tests."""
        super().setUp()
        os.environ["LANG"] = "C"
        init_i18n()


@lru_cache(1)
def running_in_ci():
    return "CI" in os.environ


def skip(reason, strict=True):
    if strict and running_in_ci():
        pytest.fail(reason)
    else:
        pytest.skip(reason)


def run(cmd, verbosity=0, **kwargs):
    """Run command without error checking.
    @return: command return code"""
    if kwargs.get("shell"):
        # for shell calls the command must be a string
        cmd = " ".join(cmd)
    return subprocess.call(cmd, **kwargs)


def run_checked(cmd, ret_ok=(0,), **kwargs):
    """Run command and raise OSError on error."""
    retcode = run(cmd, **kwargs)
    if retcode not in ret_ok:
        msg = "Command `%s' returned non-zero exit status %d" % (cmd, retcode)
        raise OSError(msg)
    return retcode


def run_silent(cmd):
    """Run given command without output."""
    null = open(os.name == "nt" and ":NUL" or "/dev/null", "w")
    try:
        return run(cmd, stdout=null, stderr=subprocess.STDOUT)
    finally:
        null.close()


def _need_func(testfunc, name, strict=True):
    """Decorator skipping test if given testfunc fails."""

    def check_func(func):
        @wraps(func)
        def newfunc(*args, **kwargs):
            if not testfunc():
                skip("%s is not available" % name, strict)
            return func(*args, **kwargs)

        return newfunc

    return check_func


@lru_cache(1)
def has_network():
    """Test if network is up."""
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(("www.python.org", 80))
        s.close()
        return True
    except Exception:
        pass
    return False


need_network = _need_func(has_network, "network")


@lru_cache(1)
def has_msgfmt():
    """Test if msgfmt is available."""
    return run_silent(["msgfmt", "-V"]) == 0


need_msgfmt = _need_func(has_msgfmt, "msgfmt")


@lru_cache(1)
def has_posix():
    """Test if this is a POSIX system."""
    return os.name == "posix"


need_posix = _need_func(has_posix, "POSIX system", False)


@lru_cache(1)
def has_windows():
    """Test if this is a Windows system."""
    return os.name == "nt"


need_windows = _need_func(has_windows, "Windows system", False)


@lru_cache(1)
def has_linux():
    """Test if this is a Linux system."""
    return sys.platform.startswith("linux")


need_linux = _need_func(has_linux, "Linux system", False)


@lru_cache(1)
def has_clamav():
    """Test if ClamAV daemon is installed and running."""
    try:
        cmd = ["grep", "LocalSocket", "/etc/clamav/clamd.conf"]
        sock = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0].split()[1]
        if sock:
            s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            s.connect(sock)
            s.close()
            return True
    except Exception:
        pass
    return False


need_clamav = _need_func(has_clamav, "ClamAV", False)  # XXX Plugin disabled


@lru_cache(1)
def has_proxy():
    """Test if proxy is running on port 8081."""
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(("localhost", 8081))
        s.close()
        return True
    except Exception:
        return False


need_proxy = _need_func(has_proxy, "proxy")


@lru_cache(1)
def has_pyftpdlib():
    """Test if pyftpdlib is available."""
    try:
        import pyftpdlib

        return True
    except ImportError:
        return False


need_pyftpdlib = _need_func(has_pyftpdlib, "pyftpdlib")


@lru_cache(1)
def has_x11():
    """Test if DISPLAY variable is set."""
    return os.getenv("DISPLAY") is not None


need_x11 = _need_func(has_x11, "X11")


@lru_cache(1)
def has_geoip():
    from linkcheck.plugins import locationinfo

    return locationinfo.geoip is not None


need_geoip = _need_func(has_geoip, "geoip")


@lru_cache(1)
def has_word():
    """Test if Word is available."""
    from linkcheck.plugins import parseword

    return parseword.has_word()


need_word = _need_func(has_word, "Word", False)


@lru_cache(1)
def has_pdflib():
    from linkcheck.plugins import parsepdf

    return parsepdf.has_pdflib


need_pdflib = _need_func(has_pdflib, "pdflib")


@contextmanager
def _limit_time(seconds):
    """Raises LinkCheckerInterrupt if given number of seconds have passed."""
    if os.name == "posix":

        def signal_handler(signum, frame):
            raise LinkCheckerInterrupt("timed out")

        old_handler = signal.getsignal(signal.SIGALRM)
        signal.signal(signal.SIGALRM, signal_handler)
        signal.alarm(seconds)
    yield
    if os.name == "posix":
        signal.alarm(0)
        if old_handler is not None:
            signal.signal(signal.SIGALRM, old_handler)


def limit_time(seconds, skip=False):
    """Limit test time to the given number of seconds, else fail or skip."""

    def run_limited(func):
        def new_func(*args, **kwargs):
            try:
                with _limit_time(seconds):
                    return func(*args, **kwargs)
            except LinkCheckerInterrupt as msg:
                if skip:
                    pytest.skip("time limit of %d seconds exceeded" % seconds)
                assert False, msg

        new_func.__name__ = func.__name__
        return new_func

    return run_limited


def get_file(filename=None):
    """
    Get file name located within 'data' directory.
    """
    directory = os.path.join("tests", "checker", "data")
    if filename:
        return os.path.join(directory, filename)
    return directory


if __name__ == "__main__":
    print("has clamav", has_clamav())
    print("has network", has_network())
    print("has msgfmt", has_msgfmt())
    print("has POSIX", has_posix())
    print("has proxy", has_proxy())
    print("has X11", has_x11())



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test html anchor parsing and checking.
"""
from . import LinkCheckTest
from .httpserver import HttpServerTest


class TestFileAnchor(LinkCheckTest):
    """
    Test anchor checking of HTML files.
    """

    def test_anchor_file(self):
        confargs = {"enabledplugins": ["AnchorCheck"]}
        anchor = "broken"
        url = "file://%(curdir)s/%(datadir)s/anchor.html#%(anchor)s" % self.get_attrs(
               anchor=anchor)
        nurl = self.norm(url)
        resultlines = [
            f"url {url}",
            f"cache key {nurl}",
            f"real url {nurl}",
            f"warning Anchor `{anchor}' (decoded: `{anchor}') not found."
            " Available anchors: `myid:'.",
            "valid",
        ]
        self.direct(url, resultlines, confargs=confargs)


class TestHttpAnchor(HttpServerTest):
    """
    Test checking of HTML pages containing links to anchors served over http.
    """

    def test_anchor_http(self):
        confargs = dict(enabledplugins=["AnchorCheck"], recursionlevel=1)
        self.file_test("http_anchor.html", confargs=confargs)


class TestEncodedAnchors(HttpServerTest):
    """Test HTML pages containing urlencoded links to anchors"""

    def test_anchor_encoded_http(self):
        """
        http://
        """
        confargs = dict(enabledplugins=["AnchorCheck"], recursionlevel=1)
        self.file_test("urlencoding_anchor.html", confargs=confargs)

    def test_anchor_encoded_file(self):
        """
        file://
        This should have identical behavior as http://
        """
        filename = "urlencoding_anchor.html"
        confargs = {"enabledplugins": ["AnchorCheck"]}
        url = "file://%(curdir)s/%(datadir)s/%(filename)s" % self.get_attrs(
               filename=filename)
        # get results from the special result file that has `.file.` in its name
        resultlines = self.get_resultlines(f"{filename}.file")
        self.direct(url, resultlines, recursionlevel=1, confargs=confargs)


class TestAnchorsAcrossMultipleFiles(LinkCheckTest):
    """Test anchors when there are multiple files"""

    def test_anchor1_file(self):
        """
        Test a network of files that reference each other, starting with anchor1.html
        """
        filename = "anchor1.html"
        confargs = {"enabledplugins": ["AnchorCheck"]}
        url = "file://%(curdir)s/%(datadir)s/%(filename)s" % self.get_attrs(
               filename=filename)
        resultlines = self.get_resultlines(filename)
        self.direct(url, resultlines, recursionlevel=4, confargs=confargs)



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Handler for mailto: links.
"""

import re
import urllib.parse
from email._parseaddr import AddressList

from . import urlbase
from .. import log, LOG_CHECK, strformat, url as urlutil
from dns import resolver
from ..network import iputil
from .const import WARN_MAIL_NO_MX_HOST


def getaddresses(addr):
    """Return list of email addresses from given field value."""
    parsed = [mail for name, mail in AddressList(addr).addresslist if mail]
    if parsed:
        addresses = parsed
    elif addr:
        # we could not parse any mail addresses, so try with the raw string
        addresses = [addr]
    else:
        addresses = []
    return addresses


def is_quoted(addr):
    """Return True iff mail address string is quoted."""
    return addr.startswith('"') and addr.endswith('"')


def is_literal(domain):
    """Return True iff domain string is a literal."""
    return domain.startswith('[') and domain.endswith(']')


_remove_quoted = re.compile(r'\\.').sub
_quotes = re.compile(r'["\\]')


def is_missing_quote(addr):
    """Return True iff mail address is not correctly quoted."""
    return _quotes.match(_remove_quoted("", addr[1:-1]))


# list of CGI keys to search for email addresses
EMAIL_CGI_ADDRESS = ("to", "cc", "bcc")
EMAIL_CGI_SUBJECT = "subject"


class MailtoUrl(urlbase.UrlBase):
    """
    Url link with mailto scheme.
    """

    def build_url(self):
        """Call super.build_url(), extract list of mail addresses from URL,
        and check their syntax.
        """
        super().build_url()
        self.addresses = set()
        self.subject = None
        self.parse_addresses()
        if self.addresses:
            for addr in sorted(self.addresses):
                self.check_email_syntax(addr)
                if not self.valid:
                    break
        elif not self.subject:
            self.add_warning(
                _("No mail addresses or email subject found in `%(url)s'.")
                % {"url": self.url}
            )

    def parse_addresses(self):
        """Parse all mail addresses out of the URL target. Also parses
        optional CGI headers like "?to=foo@example.org".
        Stores parsed addresses in the self.addresses set.
        """
        # cut off leading mailto: and unquote
        url = urllib.parse.unquote(self.base_url[7:], self.encoding)
        # search for cc, bcc, to and store in headers
        mode = 0  # 0=default, 1=quote, 2=esc
        quote = None
        i = 0
        for i, c in enumerate(url):
            if mode == 0:
                if c == '?':
                    break
                elif c in '<"':
                    quote = c
                    mode = 1
                elif c == '\\':
                    mode = 2
            elif mode == 1:
                if c == '"' and quote == '"':
                    mode = 0
                elif c == '>' and quote == '<':
                    mode = 0
            elif mode == 2:
                mode = 0
        if i < (len(url) - 1):
            self.addresses.update(getaddresses(url[:i]))
            try:
                headers = urllib.parse.parse_qs(url[(i + 1):], strict_parsing=True)
                for key, vals in headers.items():
                    if key.lower() in EMAIL_CGI_ADDRESS:
                        # Only the first header value is added
                        self.addresses.update(
                            getaddresses(urllib.parse.unquote(vals[0], self.encoding))
                        )
                    if key.lower() == EMAIL_CGI_SUBJECT:
                        self.subject = vals[0]
            except ValueError as err:
                self.add_warning(_("Error parsing CGI values: %s") % str(err))
        else:
            self.addresses.update(getaddresses(url))
        log.debug(LOG_CHECK, "addresses: %s", self.addresses)

    def check_email_syntax(self, mail):
        """Check email syntax. The relevant RFCs:

        - How to check names (memo):
          https://tools.ietf.org/html/rfc3696
        - Email address syntax
          https://tools.ietf.org/html/rfc2822
        - SMTP protocol
          https://tools.ietf.org/html/rfc5321#section-4.1.3
        - IPv6
          https://tools.ietf.org/html/rfc4291#section-2.2
        - Host syntax
          https://tools.ietf.org/html/rfc1123#section-2
        """
        # length checks

        # restrict email length to 256 characters
        # https://www.rfc-editor.org/errata_search.php?eid=1003
        if len(mail) > 256:
            self.set_result(
                _(
                    "Mail address `%(addr)s' too long. Allowed 256 chars,"
                    " was %(length)d chars."
                )
                % {"addr": mail, "length": len(mail)},
                valid=False,
                overwrite=False,
            )
            return
        if "@" not in mail:
            self.set_result(
                _("Missing `@' in mail address `%(addr)s'.") % {"addr": mail},
                valid=False,
                overwrite=False,
            )
            return
        # note: be sure to use rsplit since "@" can occur in local part
        local, domain = mail.rsplit("@", 1)
        if not local:
            self.set_result(
                _("Missing local part of mail address `%(addr)s'.") % {"addr": mail},
                valid=False,
                overwrite=False,
            )
            return
        if not domain:
            self.set_result(
                _("Missing domain part of mail address `%(addr)s'.") % {"addr": mail},
                valid=False,
                overwrite=False,
            )
            return
        if len(local) > 64:
            self.set_result(
                _(
                    "Local part of mail address `%(addr)s' too long."
                    " Allowed 64 chars, was %(length)d chars."
                )
                % {"addr": mail, "length": len(local)},
                valid=False,
                overwrite=False,
            )
            return
        if len(domain) > 255:
            self.set_result(
                _(
                    "Domain part of mail address `%(addr)s' too long."
                    " Allowed 255 chars, was %(length)d chars."
                )
                % {"addr": mail, "length": len(local)},
                valid=False,
                overwrite=False,
            )
            return

        # local part syntax check

        # Rules taken from https://tools.ietf.org/html/rfc3696#section-3
        if is_quoted(local):
            if is_missing_quote(local):
                self.set_result(
                    _("Unquoted double quote or backslash in mail address `%(addr)s'.")
                    % {"addr": mail},
                    valid=False,
                    overwrite=False,
                )
                return
        else:
            if local.startswith("."):
                self.set_result(
                    _("Local part of mail address `%(addr)s' may not start with a dot.")
                    % {"addr": mail},
                    valid=False,
                    overwrite=False,
                )
                return
            if local.endswith("."):
                self.set_result(
                    _("Local part of mail address `%(addr)s' may not end with a dot.")
                    % {"addr": mail},
                    valid=False,
                    overwrite=False,
                )
                return
            if ".." in local:
                self.set_result(
                    _("Local part of mail address `%(addr)s' may not contain two dots.")
                    % {"addr": mail},
                    valid=False,
                    overwrite=False,
                )
                return
            for char in '@ \\",[]':
                if char in local.replace(f"\\{char}", ""):
                    self.set_result(
                        _(
                            "Local part of mail address `%(addr)s' contains"
                            " unquoted character `%(char)s."
                        )
                        % {"addr": mail, "char": char},
                        valid=False,
                        overwrite=False,
                    )
                    return

        # domain part syntax check

        if is_literal(domain):
            # it's an IP address
            ip = domain[1:-1]
            if ip.startswith("IPv6:"):
                ip = ip[5:]
            if not iputil.is_valid_ip(ip):
                self.set_result(
                    _("Domain part of mail address `%(addr)s' has invalid IP.")
                    % {"addr": mail},
                    valid=False,
                    overwrite=False,
                )
                return
        else:
            # it's a domain name
            if not urlutil.is_safe_domain(domain):
                self.set_result(
                    _("Invalid domain part of mail address `%(addr)s'.")
                    % {"addr": mail},
                    valid=False,
                    overwrite=False,
                )
                return
            if domain.endswith(".") or domain.split(".")[-1].isdigit():
                self.set_result(
                    _("Invalid top level domain part of mail address `%(addr)s'.")
                    % {"addr": mail},
                    valid=False,
                    overwrite=False,
                )
                return

    def check_connection(self):
        """
        Verify a list of email addresses. If one address fails,
        the whole list will fail.

        For each mail address the MX DNS records are found.
        If no MX records are found, print a warning and try
        to look for A DNS records. If no A records are found either
        print an error.
        """
        for mail in sorted(self.addresses):
            self.check_smtp_domain(mail)
            if not self.valid:
                break

    def check_smtp_domain(self, mail):
        """
        Check a single mail address.
        """
        from dns.exception import DNSException

        log.debug(LOG_CHECK, "checking mail address %r", mail)
        mail = strformat.ascii_safe(mail)
        username, domain = mail.rsplit('@', 1)
        log.debug(LOG_CHECK, "looking up MX mailhost %r", domain)
        try:
            answers = resolver.resolve(domain, 'MX', search=True)
        except DNSException:
            answers = []
        if len(answers) == 0:
            self.add_warning(
                _("No MX mail host for %(domain)s found.") % {'domain': domain},
                tag=WARN_MAIL_NO_MX_HOST,
            )
            try:
                answers = resolver.resolve(domain, 'A', search=True)
            except DNSException:
                answers = []
            if len(answers) == 0:
                self.set_result(
                    _("No host for %(domain)s found.") % {'domain': domain},
                    valid=False,
                    overwrite=True,
                )
                return
            # set preference to zero
            mxdata = [(0, rdata.to_text(omit_final_dot=True)) for rdata in answers]
        else:
            from dns.rdtypes.mxbase import MXBase

            mxdata = [
                (rdata.preference, rdata.exchange.to_text(omit_final_dot=True))
                for rdata in answers
                if isinstance(rdata, MXBase)
            ]
            if not mxdata:
                self.set_result(
                    _("Got invalid DNS answer %(answer)s for %(domain)s.")
                    % {'answer': answers, 'domain': domain},
                    valid=False,
                    overwrite=True,
                )
                return
            # sort according to preference (lower preference means this
            # host should be preferred)
            mxdata.sort()
        # debug output
        log.debug(LOG_CHECK, "found %d MX mailhosts:", len(answers))
        for preference, host in mxdata:
            log.debug(LOG_CHECK, "MX host %r, preference %d", host, preference)
        self.set_result(_("Valid mail address syntax"))

    def set_cache_url(self):
        """
        The cache url is a comma separated list of emails.
        """
        emails = ",".join(sorted(self.addresses))
        self.cache_url = f"{self.scheme}:{emails}"

    def can_get_content(self):
        """
        mailto: URLs do not have any content

        @return: False
        @rtype: bool
        """
        return False



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test html <base> tag parsing.
"""
from . import LinkCheckTest


class TestBase(LinkCheckTest):
    """
    Test, if charset encoding is done right.
    The linkchecker should translate the encoding
    from the original source and show it on the terminal
    in most readable form.
    Check the tested files with browser - the link text
    should look the same in all of them.
    """

    def test_utf8(self):
        self.file_test("charsets/utf8.html")

    def test_iso8859_2(self):
        self.file_test("charsets/iso8859-2.html")

    def test_cp1250(self):
        self.file_test("charsets/cp1250.html")



# Copyright (C) 2000-2016 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Create command line arguments.
"""

import argparse

from .. import checker, logconf, logger, COMMAND_NAME

from ..cmdline import LCArgumentParser

# usage texts
Notes = _(
    """NOTES
 o URLs on the command line starting with "ftp." are treated like
   "ftp://ftp.", URLs starting with "www." are treated like "http://www.".
   You can also give local files as arguments.
 o If you have your system configured to automatically establish a
   connection to the internet (e.g. with diald), it will connect when
   checking links not pointing to your local system.
   See the --ignore-url option on how to prevent this.
 o Javascript links are currently ignored.
 o If your platform does not support threading, LinkChecker disables it
   automatically.
 o You can supply multiple user/password pairs in a configuration file.
"""
)

ProxySupport = _(
    """PROXY SUPPORT
To use a proxy on Unix or Windows set $http_proxy or $https_proxy
to the proxy URL. The URL should be of the form
"http://[<user>:<pass>@]<host>[:<port>]".
LinkChecker also detects manual proxy settings of Internet Explorer under
Windows systems. On a Mac use the Internet Config to select a proxy.

LinkChecker honors the $no_proxy environment variable. It can be a list
of domain names for which no proxy will be used.

Setting a HTTP proxy on Unix for example looks like this:

  export http_proxy="http://proxy.example.com:8080"

Proxy authentication is also supported:

  export http_proxy="http://user1:mypass@proxy.example.org:8081"

Setting a proxy on the Windows command prompt:

  set http_proxy=http://proxy.example.com:8080

"""
)

RegularExpressions = _(
    """REGULAR EXPRESSIONS
Only Python regular expressions are accepted by LinkChecker.
See https://docs.python.org/howto/regex.html for an introduction in
regular expressions.

The only addition is that a leading exclamation mark negates
the regular expression.
"""
)

CookieFormat = _(
    """COOKIE FILES
A cookie file contains standard RFC 805 header data with the following
possible names:
Scheme (optional)
 Sets the scheme the cookies are valid for; default scheme is 'http'.
Host (required)
 Sets the domain the cookies are valid for.
Path (optional)
 Gives the path the cookies are value for; default path is '/'.
Set-cookie (optional)
 Set cookie name/value. Can be given more than once.

Multiple entries are separated by a blank line.

The example below will send two cookies to all URLs starting with
'http://example.org/hello/' and one to all URLs starting
with 'https://example.com/':

Host: example.org
Path: /hello
Set-cookie: ID="smee"
Set-cookie: spam="egg"

Scheme: https
Host: example.com
Set-cookie: baggage="elitist"; comment="hologram"
"""
)

Retval = _(
    r"""RETURN VALUE
The return value is non-zero when
 o invalid links were found or
 o warnings were found warnings are enabled
 o a program error occurred
"""
)

Examples = _(
    r"""EXAMPLES
The most common use checks the given domain recursively, plus any
single URL pointing outside of the domain:
  linkchecker http://www.example.org/
Beware that this checks the whole site which can have several hundred
thousands URLs. Use the -r option to restrict the recursion depth.

Don't connect to mailto: hosts, only check their URL syntax. All other
links are checked as usual:
  linkchecker --ignore-url=^mailto: www.example.org

Checking local HTML files on Unix:
  linkchecker ../bla.html subdir/blubber.html

Checking a local HTML file on Windows:
  linkchecker c:\temp\test.html

You can skip the "http://" url part if the domain starts with "www.":
  linkchecker www.example.de

You can skip the "ftp://" url part if the domain starts with "ftp.":
  linkchecker -r0 ftp.example.org
"""
)

LoggerTypes = _(
    r"""OUTPUT TYPES
Note that by default only errors and warnings are logged.
You should use the --verbose option to see valid URLs,
and when outputting a sitemap graph format.

text    Standard text output, logging URLs in keyword: argument fashion.
html    Log URLs in keyword: argument fashion, formatted as HTML.
        Additionally has links to the referenced pages. Invalid URLs have
        HTML and CSS syntax check links appended.
csv     Log check result in CSV format with one URL per line.
gml     Log parent-child relations between linked URLs as a GML sitemap
        graph.
dot     Log parent-child relations between linked URLs as a DOT sitemap
        graph.
gxml    Log check result as a GraphXML sitemap graph.
xml     Log check result as machine-readable XML.
sql     Log check result as SQL script with INSERT commands. An example
        script to create the initial SQL table is included as create.sql.
failures
        Suitable for cron jobs. Logs the check result into a file
        $XDG_DATA_HOME/linkchecker/failures which only contains entries with
        invalid URLs and the number of times they have failed.
none    Logs nothing. Suitable for debugging or checking the exit code.
"""
)

Warnings = (
    _(
        r"""IGNORE WARNINGS
The following warnings are recognized in the 'ignorewarnings' config
file entry:
"""
    )
    + "\n".join(
        [
            f" o {tag} - {desc}"
            for tag, desc in sorted(checker.const.Warnings.items())
        ]
    )
)

Epilog = "\n".join(
    (
        Examples,
        LoggerTypes,
        RegularExpressions,
        CookieFormat,
        ProxySupport,
        Notes,
        Retval,
        Warnings,
    )
)


class ArgParser(LCArgumentParser):
    """Create a parser for command line arguments"""

    def __init__(self):
        super().__init__(
            epilog=Epilog,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            prog=COMMAND_NAME,
        )

        # ================== general options =====================
        group = self.add_argument_group(_("General options"))
        group.add_argument(
            "-f",
            "--config",
            dest="configfile",
            metavar="FILENAME",
            help=_(
                "Use FILENAME as configuration file. Per default LinkChecker uses\n"
                "$XDG_CONFIG_HOME/linkchecker/linkcheckerrc (under Windows\n"
                "%%HOMEPATH%%\\.config\\linkchecker\\linkcheckerrc)."
            ),
        )
        group.add_argument(
            "-t",
            "--threads",
            type=int,
            metavar="NUMBER",
            help=_(
                "Generate no more than the given number of threads. Default number\n"
                "of threads is 10. To disable threading specify a non-positive number."
            ),
        )
        group.add_argument(
            "-V", "--version", action="store_true", help=_("Print version and exit.")
        )
        group.add_argument(
            "--list-plugins",
            action="store_true",
            dest="listplugins",
            help=_("Print available check plugins and exit."),
        )
        group.add_argument(
            "--stdin",
            action="store_true",
            help=_("Read list of white-space separated URLs to check from stdin."),
        )

        # ================== output options =====================
        group = self.add_argument_group(_("Output options"))
        group.add_argument(
            "-D",
            "--debug",
            action="append",
            metavar="STRING",
            help=_(
                "Print debugging output for the given logger.\n"
                "Available loggers are %(lognamelist)s.\n"
                "Specifying 'all' is an alias for specifying all available loggers.\n"
                "The option can be given multiple times to debug with more\n"
                "than one logger.\n"
                "\n"
                "For accurate results, threading will be disabled during debug runs."
            )
            % {"lognamelist": logconf.lognamelist},
        )
        group.add_argument(
            "-F",
            "--file-output",
            action="append",
            dest="fileoutput",
            metavar="TYPE[/ENCODING[/FILENAME]]",
            help=_(
                "Output to a file linkchecker-out.TYPE, $XDG_DATA_HOME/linkchecker/failures for\n"
                "'failures' output, or FILENAME if specified.\n"
                "The ENCODING specifies the output encoding, the default is that of your\n"
                "locale.\n"
                "Valid encodings are listed at "
                "https://docs.python.org/library/codecs.html#standard-encodings.\n"
                "The FILENAME and ENCODING parts of the 'none' output type will be ignored,\n"
                "else if the file already exists, it will be overwritten.\n"
                "You can specify this option more than once. Valid file output types\n"
                "are %(loggertypes)s. You can specify this option multiple times to output\n"
                "to more than one file. Default is no file output. Note that you can\n"
                "suppress all console output with the option '-o none'."
            )
            % {"loggertypes": logger.LoggerKeys},
        )
        group.add_argument(
            "--no-status",
            action="store_false",
            default=None,
            dest="status",
            help=_("Do not print check status messages."),
        )
        group.add_argument(
            "--no-warnings",
            action="store_false",
            dest="warnings",
            help=_("Don't log warnings. Default is to log warnings."),
        )
        group.add_argument(
            "-o",
            "--output",
            dest="output",
            metavar="TYPE[/ENCODING]",
            help=_(
                "Specify output as %(loggertypes)s. Default output type is text.\n"
                "The ENCODING specifies the output encoding, the default is that of your\n"
                "locale.\n"
                "Valid encodings are listed at "
                "https://docs.python.org/library/codecs.html#standard-encodings."
            )
            % {"loggertypes": logger.LoggerKeys},
        )
        group.add_argument(
            "--profile", action="store_true", dest="profile", help=argparse.SUPPRESS
        )
        group.add_argument(
            "-q",
            "--quiet",
            action="store_true",
            dest="quiet",
            help=_(
                "Quiet operation, an alias for '-o none'.\n"
                "This is only useful with -F."
            ),
        )
        group.add_argument(
            "--trace", action="store_true", dest="trace", help=argparse.SUPPRESS)
        group.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            dest="verbose",
            help=_("Log all URLs. Default is to log only errors and warnings."),
        )

        # =================== checking options ====================
        group = self.add_argument_group(_("Checking options"))
        group.add_argument(
            "--cookiefile",
            dest="cookiefile",
            metavar="FILENAME",
            help=_(
                "Read a file with initial cookie data. The cookie data format is\n"
                "explained below."
            ),
        )
        # const because store_false doesn't detect absent flags
        group.add_argument(
            "--no-robots",
            action="store_const",
            const=False,
            dest="norobotstxt",
            help=_("Disable robots.txt checks"),
        )
        group.add_argument(
            "--check-extern",
            action="store_true",
            dest="checkextern",
            help=_("""Check external URLs."""),
        )
        group.add_argument(
            "--ignore-url",
            action="append",
            metavar="REGEX",
            dest="externstrict",
            help=_(
                "Only check syntax of URLs matching the given regular expression.\n"
                "This option can be given multiple times."
            ),
        )
        group.add_argument(
            "--no-follow-url",
            action="append",
            metavar="REGEX",
            dest="extern",
            help=_(
                "Check but do not recurse into URLs matching the given regular\n"
                "expression. This option can be given multiple times."
            ),
        )
        group.add_argument(
            "-p",
            "--password",
            action="store_true",
            dest="password",
            help=_(
                "Read a password from console and use it for HTTP and FTP authorization.\n"
                "For FTP the default password is 'anonymous@'. For HTTP there is\n"
                "no default password. See also -u."
            ),
        )
        group.add_argument(
            "-r",
            "--recursion-level",
            type=int,
            dest="recursionlevel",
            metavar="NUMBER",
            help=_(
                "Check recursively all links up to given depth. A negative depth\n"
                "will enable infinite recursion. Default depth is infinite."
            ),
        )
        group.add_argument(
            "--timeout",
            type=int,
            dest="timeout",
            metavar="NUMBER",
            help=_(
                "Set the timeout for connection attempts in seconds."
            )
        )
        group.add_argument(
            "-u",
            "--user",
            dest="username",
            metavar="STRING",
            help=_(
                "Try the given username for HTTP and FTP authorization.\n"
                "For FTP the default username is 'anonymous'. For HTTP there is\n"
                "no default username. See also -p."
            ),
        )
        group.add_argument(
            "--user-agent",
            dest="useragent",
            metavar="STRING",
            help=_(
                "Specify the User-Agent string to send to the HTTP server, for example\n"
                '"Mozilla/4.0". The default is "LinkChecker/X.Y" where X.Y is the current\n'
                "version of LinkChecker."
            ),
        )

        self.add_argument("url", nargs="*")



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Base class for XML loggers.
"""

import xml.sax.saxutils
from . import _Logger


xmlattr_entities = {
    "&": "&amp;",
    "<": "&lt;",
    ">": "&gt;",
    "\"": "&quot;",
}


def xmlquote(s):
    """
    Quote characters for XML.
    """
    return xml.sax.saxutils.escape(s)


def xmlquoteattr(s):
    """
    Quote XML attribute, ready for inclusion with double quotes.
    """
    return xml.sax.saxutils.escape(s, xmlattr_entities)


class _XMLLogger(_Logger):
    """Base class for XML output; easy to parse with any XML tool."""

    def __init__(self, **kwargs):
        """ Initialize graph node list and internal id counter. """
        args = self.get_args(kwargs)
        super().__init__(**args)
        self.init_fileoutput(args)
        self.indent = "  "
        self.level = 0

    def comment(self, s, **args):
        """
        Write XML comment.
        """
        self.write("<!-- ")
        self.write(s, **args)
        self.writeln(" -->")

    def xml_start_output(self):
        """
        Write start of checking info as xml comment.
        """
        self.writeln(
            '<?xml version="1.0" encoding="%s"?>'
            % xmlquoteattr(self.get_charset_encoding())
        )
        if self.has_part("intro"):
            self.write_intro()
            self.writeln()

    def xml_end_output(self):
        """
        Write end of checking info as xml comment.
        """
        if self.has_part("outro"):
            self.write_outro()

    def xml_starttag(self, name, attrs=None):
        """
        Write XML start tag.
        """
        self.write(self.indent * self.level)
        self.write("<%s" % xmlquote(name))
        if attrs:
            for name, value in attrs.items():
                args = (xmlquote(name), xmlquoteattr(value))
                self.write(' %s="%s"' % args)
        self.writeln(">")
        self.level += 1

    def xml_endtag(self, name):
        """
        Write XML end tag.
        """
        self.level -= 1
        assert self.level >= 0
        self.write(self.indent * self.level)
        self.writeln("</%s>" % xmlquote(name))

    def xml_tag(self, name, content, attrs=None):
        """
        Write XML tag with content.
        """
        self.write(self.indent * self.level)
        self.write("<%s" % xmlquote(name))
        if attrs:
            for aname, avalue in attrs.items():
                args = (xmlquote(aname), xmlquoteattr(avalue))
                self.write(' %s="%s"' % args)
        self.writeln(f">{xmlquote(content)}</{xmlquote(name)}>")



# Copyright (C) 2009-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Python implementation of a part of Dan Bernstein's ftpparse library.

See also http://cr.yp.to/ftpparse.html
"""

months = (
    "jan",
    "feb",
    "mar",
    "apr",
    "may",
    "jun",
    "jul",
    "aug",
    "sep",
    "oct",
    "nov",
    "dec",
)


def ismonth(txt):
    """Check if given text is a month name."""
    return txt.lower() in months


def ftpparse(line):
    """Parse a FTP list line into a dictionary with attributes:
    name - name of file (string)
    trycwd - False if cwd is definitely pointless, True otherwise
    tryretr - False if retr is definitely pointless, True otherwise

    If the line has no file information, None is returned
    """
    if len(line) < 2:
        # an empty name in EPLF, with no info, could be 2 chars
        return None
    info = dict(name=None, trycwd=False, tryretr=False)

    # EPLF format
    # http://pobox.com/~djb/proto/eplf.html
    # "+i8388621.29609,m824255902,/,\tdev"
    # "+i8388621.44468,m839956783,r,s10376,\tRFCEPLF"
    if line[0] == '+':
        if '\t' in line:
            flags, name = line.split('\t', 1)
            info['name'] = name
            flags = flags.split(',')
            info['trycwd'] = '/' in flags
            info['tryretr'] = 'r' in flags
        return info

    # UNIX-style listing, without inum and without blocks
    # "-rw-r--r--   1 root     other        531 Jan 29 03:26 README"
    # "dr-xr-xr-x   2 root     other        512 Apr  8  1994 etc"
    # "dr-xr-xr-x   2 root     512 Apr  8  1994 etc"
    # "lrwxrwxrwx   1 root     other          7 Jan 25 00:17 bin -> usr/bin"
    # Also produced by Microsoft's FTP servers for Windows:
    # "----------   1 owner    group         1803128 Jul 10 10:18 ls-lR.Z"
    # "d---------   1 owner    group               0 May  9 19:45 Softlib"
    # Also WFTPD for MS-DOS:
    # "-rwxrwxrwx   1 noone    nogroup      322 Aug 19  1996 message.ftp"
    # Also NetWare:
    # "d [R----F--] supervisor            512       Jan 16 18:53    login"
    # "- [R----F--] rhesus             214059       Oct 20 15:27    cx.exe"
    # Also NetPresenz for the Mac:
    # "-------r--         326  1391972  1392298 Nov 22  1995 MegaPhone.sit"
    # "drwxrwxr-x               folder        2 May 10  1996 network"
    if line[0] in 'bcdlps-':
        if line[0] == 'd':
            info['trycwd'] = True
        if line[0] == '-':
            info['tryretr'] = True
        if line[0] == 'l':
            info['trycwd'] = info['tryretr'] = True
        parts = line.split()
        if len(parts) < 7:
            return None
        del parts[0]  # skip permissions
        if parts[0] != 'folder':
            del parts[0]  # skip nlink
        del parts[0]  # skip uid
        del parts[0]  # skip gid or size
        if not ismonth(parts[0]):
            del parts[0]  # skip size
        if not ismonth(parts[0]):
            return None
        del parts[0]  # skip month
        del parts[0]  # skip day
        if not parts:
            return None
        del parts[0]  # skip year or time
        name = " ".join(parts)
        # resolve links
        if line[0] == 'l' and ' -> ' in name:
            name = name.split(' -> ', 1)[1]
        # eliminate extra NetWare spaces
        if line[1] in ' [' and name.startswith('   '):
            name = name[3:]
        info["name"] = name
        return info

    # MultiNet (some spaces removed from examples)
    # "00README.TXT;1      2 30-DEC-1996 17:44 [SYSTEM] (RWED,RWED,RE,RE)"
    # "CORE.DIR;1          1  8-SEP-1996 16:09 [SYSTEM] (RWE,RWE,RE,RE)"
    # and non-MutliNet VMS:
    # "CII-MANUAL.TEX;1  213/216  29-JAN-1996 03:33:12  [ANONYMOU,ANONYMOUS]   (RWED,RWED,,)"
    i = line.find(';')
    if i != -1:
        name = line[:i]
        if name.endswith(".DIR"):
            name = name[:-4]
            info["trycwd"] = True
        else:
            info["tryretr"] = True
        info["name"] = name
        return info

    # MS-DOS format
    # 04-27-00  09:09PM       <DIR>          licensed
    # 07-18-00  10:16AM       <DIR>          pub
    # 04-14-00  03:47PM                  589 readme.htm
    if line[0].isdigit():
        parts = line.split()
        if len(parts) != 4:
            return None
        info['name'] = parts[3]
        if parts[2][0] == '<':
            info['trycwd'] = True
        else:
            info['tryretr'] = True
        return info

    # Some useless lines, safely ignored:
    # "Total of 11 Files, 10966 Blocks." (VMS)
    # "total 14786" (UNIX)
    # "DISK$ANONFTP:[ANONYMOUS]" (VMS)
    # "Directory DISK$PCSA:[ANONYM]" (VMS)
    return None



# Copyright (C) 2010-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test file utility functions.
"""

import unittest
import linkcheck.fileutil

file_existing = __file__
file_non_existing = "ZZZ.i_dont_exist"


class TestFileutil(unittest.TestCase):
    """Test file utility functions."""

    def test_size(self):
        self.assertTrue(linkcheck.fileutil.get_size(file_existing) > 0)
        self.assertEqual(linkcheck.fileutil.get_size(file_non_existing), -1)

    def test_mtime(self):
        self.assertTrue(linkcheck.fileutil.get_mtime(file_existing) > 0)
        self.assertEqual(linkcheck.fileutil.get_mtime(file_non_existing), 0)



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test http checking.
"""
from tests import need_network
from .httpserver import HttpServerTest, CookieRedirectHttpRequestHandler


class TestHttpRedirect(HttpServerTest):
    """Test http:// link redirection checking."""

    def __init__(self, methodName="runTest"):
        super().__init__(methodName=methodName)
        self.handler = CookieRedirectHttpRequestHandler

    @need_network
    def test_redirect(self):
        self.redirect1()
        self.redirect2()
        self.redirect3()
        self.redirect4()
        self.redirect5()

    def redirect1(self):
        url = "http://localhost:%d/redirect1" % self.port
        nurl = url
        rurl = url.replace("redirect", "newurl")
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % rurl,
            "warning Redirected to `%s' status: 302 Found." % rurl,
            "error",
        ]
        self.direct(url, resultlines, recursionlevel=0)

    def redirect2(self):
        url = self.get_url("redirect.html")
        nurl = url
        rurl = url.replace("redirect", "newurl")
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % rurl,
            "warning Redirected to `%s' status: 302 Found." % rurl,
            "valid",
        ]
        self.direct(url, resultlines, recursionlevel=99)

    def redirect3(self):
        url = "http://localhost:%d/tests/checker/data/redir.html" % self.port
        resultlines = self.get_resultlines("redir.html")
        self.direct(url, resultlines, recursionlevel=1)

    def redirect4(self):
        url = "http://localhost:%d/redirect_newscheme_ftp" % self.port
        nurl = url
        # rurl = "ftp://example.com/"
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            # don't allow ftp redirects
            # "info Redirected to `%s'." % rurl,
            # "valid",
            # "url %s" % rurl,
            # "cache key %s" % rurl,
            # "real url %s" % rurl,
            "error",
        ]
        self.direct(url, resultlines, recursionlevel=99)

    def redirect5(self):
        url = "http://localhost:%d/redirect_newscheme_file" % self.port
        nurl = url
        # rurl = "file:README"
        # rnurl = "file:///README"
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            # don't allow file redirects
            # "info Redirected to `%s'." % rurl,
            # "warning Redirection to url `%s' is not allowed." % rnurl,
            "error",
        ]
        self.direct(url, resultlines, recursionlevel=99)

    def redirect6(self):
        # max_redirect = 10
        # url = "http://httpbin.org/redirect/" + max_redirect --> valid
        # url = "http://httpbin.org/redirect/" + (max_redirect+1) --> error
        pass  # XXX



# Copyright (C) 2011 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test console operations.
"""

import linkcheck.director.console

from . import TestBase


class TestConsole(TestBase):
    """Test console operations."""

    def test_internal_error(self):
        linkcheck.director.console.internal_error()



# Copyright (C) 2006-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
import _thread

from ..decorators import notimplemented
from .. import threader
from . import console


class CheckedTask(threader.StoppableThread):
    """Stoppable URL check task, handling error conditions while running."""

    def run(self):
        """Handle keyboard interrupt and other errors."""
        try:
            self.run_checked()
        except KeyboardInterrupt:
            _thread.interrupt_main()
        except Exception:
            self.internal_error()

    @notimplemented
    def run_checked(self):
        """Overload in subclass."""
        pass

    @notimplemented
    def internal_error(self):
        """Overload in subclass."""
        pass


class LoggedCheckedTask(CheckedTask):
    """URL check task with a logger instance and internal error handling."""

    def __init__(self, logger):
        """Initialize super instance and store given logger."""
        super().__init__()
        self.logger = logger

    def internal_error(self):
        """Log an internal error on console and the logger."""
        console.internal_error()
        self.logger.log_internal_error()



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test ignoring of warnings.
"""

from re import compile as re_compile

from . import LinkCheckTest


class TestIgnoreWarnings(LinkCheckTest):
    """
    Test whether ignoring of warnings per URL works.
    """

    def test_given_warning_for_given_url_ignored(self):
        confargs = {
            "ignorewarningsforurls": [
                (re_compile("test.txt"), re_compile("url-content-size-zero"))
            ]
        }
        self.file_test("base_ignorewarnings.html", confargs=confargs)

    def test_warning_for_unmatching_url_not_ignored(self):
        confargs = {
            "ignorewarningsforurls": [
                (re_compile("test_incorrect.txt"), re_compile("url-content-size-zero"))
            ]
        }
        self.file_test("base_ignorewarnings_with_warning.html", confargs=confargs)

    def test_non_matching_warning_for_matching_url_not_ignored(self):
        confargs = {
            "ignorewarningsforurls": [
                (re_compile("test.txt"), re_compile("not-a-warning"))
            ]
        }
        self.file_test("base_ignorewarnings_with_warning.html", confargs=confargs)

    def test_empty_warning_spec_matches_anything(self):
        confargs = {
            "ignorewarningsforurls": [
                (re_compile("test.txt"), re_compile(""))
            ]
        }
        self.file_test("base_ignorewarnings.html", confargs=confargs)



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test mail checking.
"""
from tests import need_network
from . import MailTest


class TestMailGood(MailTest):
    """
    Test mailto: link checking.
    """

    @need_network
    def test_good_mail(self):
        # some good mailto addrs
        url = self.norm(
            "mailto:Dude <calvin@users.sourceforge.net> , "
            "Killer <calvin@users.sourceforge.net>?subject=bla"
        )
        resultlines = [
            "url %s" % url,
            "cache key mailto:calvin@users.sourceforge.net",
            "real url %s" % url,
            "valid",
        ]
        self.direct(url, resultlines)
        url = self.norm(
            "mailto:Bastian Kleineidam <calvin@users.sourceforge.net>?"
            "bcc=calvin%40users.sourceforge.net"
        )
        resultlines = [
            "url %s" % url,
            "cache key mailto:calvin@users.sourceforge.net",
            "real url %s" % url,
            "valid",
        ]
        self.direct(url, resultlines)
        url = self.norm("mailto:Bastian Kleineidam <calvin@users.sourceforge.net>")
        resultlines = [
            "url %s" % url,
            "cache key mailto:calvin@users.sourceforge.net",
            "real url %s" % url,
            "valid",
        ]
        self.direct(url, resultlines)
        url = self.norm("mailto:o'hara@users.sourceforge.net")
        resultlines = [
            "url %s" % url,
            "cache key mailto:o'hara@users.sourceforge.net",
            "real url %s" % url,
            "valid",
        ]
        self.direct(url, resultlines)
        url = self.norm(
            "mailto:?to=calvin@users.sourceforge.net&subject=blubb&"
            "cc=calvin_cc@users.sourceforge.net&CC=calvin_CC@users.sourceforge.net"
        )
        resultlines = [
            "url %s" % url,
            "cache key mailto:calvin@users.sourceforge.net,"
            "calvin_CC@users.sourceforge.net,calvin_cc@users.sourceforge.net",
            "real url %s" % url,
            "valid",
        ]
        self.direct(url, resultlines)
        url = self.norm(
            "mailto:news-admins@freshcode.club?subject="
            "Re:%20[fm%20#11093]%20(news-admins)%20Submission%20"
            "report%20-%20Pretty%20CoLoRs"
        )
        resultlines = [
            "url %s" % url,
            "cache key mailto:news-admins@freshcode.club",
            "real url %s" % url,
            "valid",
        ]
        self.direct(url, resultlines)

    @need_network
    def test_warn_mail(self):
        # some mailto addrs with warnings
        # contains non-quoted characters
        url = "mailto:calvin@users.sourceforge.net?subject=\xe4\xf6\xfc"
        qurl = self.norm(url, encoding="iso-8859-1")
        resultlines = [
            "url %s" % url,
            "cache key mailto:calvin@users.sourceforge.net",
            "real url %s" % qurl,
            "valid",
        ]
        self.direct(url, resultlines, url_encoding="iso-8859-1")
        url = "mailto:calvin@users.sourceforge.net?subject=Halli hallo"
        qurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key mailto:calvin@users.sourceforge.net",
            "real url %s" % qurl,
            "valid",
        ]
        self.direct(url, resultlines)
        url = "mailto:"
        resultlines = [
            "url %s" % url,
            "cache key mailto:",
            "real url %s" % url,
            "warning No mail addresses or email subject found in `%s'." % url,
            "valid",
        ]
        self.direct(url, resultlines)

    def _mail_valid_unverified(self, char):
        # valid mail addresses
        addr = "abc%sdef@sourceforge.net" % char
        url = "mailto:%s" % addr
        self.mail_valid(url, cache_key=url)

    @need_network
    def test_valid_mail1(self):
        for char in "!#$&'":
            self._mail_valid_unverified(char)

    @need_network
    def test_valid_mail2(self):
        for char in "*+-/=":
            self._mail_valid_unverified(char)

    @need_network
    def test_valid_mail3(self):
        for char in "^_`.":
            self._mail_valid_unverified(char)

    @need_network
    def test_valid_mail4(self):
        for char in "{|}~":
            self._mail_valid_unverified(char)

    @need_network
    def test_unicode_mail(self):
        mailto = "mailto:\xf6lvin@users.sourceforge.net"
        url = self.norm(mailto, encoding="iso-8859-1")
        resultlines = [
            "url %s" % mailto,
            "cache key %s" % mailto,
            "real url %s" % url,
            "valid",
        ]
        self.direct(mailto, resultlines, url_encoding="iso-8859-1")

    @need_network
    def test_mail_subject(self):
        url = "mailto:?subject=Halli hallo"
        nurl = self.norm(url)
        curl = "mailto:"
        resultlines = [
            "url %s" % url,
            "cache key %s" % curl,
            "real url %s" % nurl,
            "valid",
        ]
        self.direct(url, resultlines)



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Handle FTP links.
"""

import ftplib
from io import BytesIO

from .. import log, LOG_CHECK, LinkCheckerError, mimeutil
from . import internpaturl, get_index_html
from .const import WARN_FTP_MISSING_SLASH


class FtpUrl(internpaturl.InternPatternUrl):
    """
    Url link with ftp scheme.
    """

    def reset(self):
        """
        Initialize FTP url data.
        """
        super().reset()
        # list of files for recursion
        self.files = []
        # last part of URL filename
        self.filename = None
        self.filename_encoding = 'iso-8859-1'

    def check_connection(self):
        """
        Check in this order: login, changing directory, list the file.
        """
        self.login()
        self.negotiate_encoding()
        self.filename = self.cwd()
        self.listfile()
        self.files = []
        return None

    def login(self):
        """Log into ftp server and check the welcome message."""
        self.url_connection = ftplib.FTP(timeout=self.aggregate.config["timeout"])
        if log.is_debug(LOG_CHECK):
            self.url_connection.set_debuglevel(1)
        try:
            self.url_connection.connect(self.host, self.port)
            _user, _password = self.get_user_password()
            if _user is None:
                self.url_connection.login()
            elif _password is None:
                self.url_connection.login(_user)
            else:
                self.url_connection.login(_user, _password)
            info = self.url_connection.getwelcome()
            if info:
                # note that the info may change every time a user logs in,
                # so don't add it to the url_data info.
                log.debug(LOG_CHECK, "FTP info %s", info)
            else:
                raise LinkCheckerError(_("Got no answer from FTP server"))
        except EOFError as msg:
            raise LinkCheckerError(
                _("Remote host has closed connection: %(msg)s") % str(msg)
            )

    def negotiate_encoding(self):
        """Check if server can handle UTF-8 encoded filenames.
        See also RFC 2640."""
        try:
            features = self.url_connection.sendcmd("FEAT")
        except ftplib.error_perm as msg:
            log.debug(LOG_CHECK, "Ignoring error when getting FTP features: %s" % msg)
        else:
            log.debug(LOG_CHECK, "FTP features %s", features)
            if " UTF-8" in features.splitlines():
                self.filename_encoding = "utf-8"

    def cwd(self):
        """
        Change to URL parent directory. Return filename of last path
        component.
        """
        path = self.urlparts[2]
        if isinstance(path, bytes):
            path = path.decode(self.filename_encoding, 'replace')
        dirname = path.strip('/')
        dirs = dirname.split('/')
        filename = dirs.pop()
        self.url_connection.cwd('/')
        for d in dirs:
            self.url_connection.cwd(d)
        return filename

    def listfile(self):
        """
        See if filename is in the current FTP directory.
        """
        if not self.filename:
            return
        files = self.get_files()
        log.debug(LOG_CHECK, "FTP files %s", str(files))
        if self.filename in files:
            # file found
            return
        # it could be a directory if the trailing slash was forgotten
        if f"{self.filename}/" in files:
            if not self.url.endswith('/'):
                self.add_warning(
                    _("Missing trailing directory slash in ftp url."),
                    tag=WARN_FTP_MISSING_SLASH,
                )
                self.url += '/'
            return
        raise ftplib.error_perm("550 File not found")

    def get_files(self):
        """Get list of filenames in directory. Subdirectories have an
        ending slash."""
        files = []

        def add_entry(line):
            """Parse list line and add the entry it points to to the file
            list."""
            log.debug(LOG_CHECK, "Directory entry %r", line)
            from ..ftpparse import ftpparse

            fpo = ftpparse(line)
            if fpo is not None and fpo["name"]:
                name = fpo["name"]
                if fpo["trycwd"]:
                    name += "/"
                if fpo["trycwd"] or fpo["tryretr"]:
                    files.append(name)

        self.url_connection.dir(add_entry)
        return files

    def is_parseable(self):
        """See if URL target is parseable for recursion."""
        if self.is_directory():
            return True
        return self.is_content_type_parseable()

    def is_directory(self):
        """See if URL target is a directory."""
        # either the path is empty, or ends with a slash
        path = self.urlparts[2]
        return (not path) or path.endswith('/')

    def set_content_type(self):
        """Set URL content type, or an empty string if content
        type could not be found."""
        self.content_type = mimeutil.guess_mimetype(self.url, read=self.get_content)
        log.debug(LOG_CHECK, "MIME type: %s", self.content_type)

    def read_content(self):
        """Return URL target content, or in case of directories a dummy HTML
        file with links to the files."""
        if self.is_directory():
            self.url_connection.cwd(self.filename)
            self.files = self.get_files()
            # XXX limit number of files?
            data = get_index_html(self.files)
        else:
            # download file in BINARY mode
            ftpcmd = f"RETR {self.filename}"
            buf = BytesIO()

            def stor_data(s):
                """Helper method storing given data"""
                # limit the download size
                if (buf.tell() + len(s)) > self.aggregate.config["maxfilesizedownload"]:
                    raise LinkCheckerError(_("FTP file size too large"))
                buf.write(s)

            self.url_connection.retrbinary(ftpcmd, stor_data)
            data = buf.getvalue()
            buf.close()
        return data

    def close_connection(self):
        """Release the open connection from the connection pool."""
        if self.url_connection is not None:
            try:
                self.url_connection.quit()
            except Exception:
                pass
            self.url_connection = None



# Copyright (C) 2005-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Parsing of cookies.
"""

from http.cookiejar import split_header_words
import email
import requests

from . import LinkCheckerError


def from_file(filename):
    """Parse cookie data from a text file in HTTP header format.

    @return: list of tuples (headers, scheme, host, path)
    """
    entries = []
    with open(filename) as fd:
        lines = []
        for line in fd.readlines():
            line = line.rstrip()
            if not line:
                if lines:
                    entries.extend(from_headers("\r\n".join(lines)))
                lines = []
            else:
                lines.append(line)
        if lines:
            entries.extend(from_headers("\r\n".join(lines)))
        if not entries:
            raise LinkCheckerError(_("No entries found"))
        return entries


def from_headers(strheader):
    """Parse cookie data from a string in HTTP header (RFC 2616) format.

    @return: list of cookies
    @raises: ValueError for incomplete or invalid data
    """
    res = []
    headers = email.message_from_string(strheader)
    if "Host" not in headers:
        raise ValueError("Required header 'Host:' missing")
    host = headers["Host"]
    # XXX: our --help says we also pay attention to the Scheme: header,
    # but we don't?!
    path = headers.get("Path", "/")
    for headervalue in headers.get_all("Set-Cookie"):
        for pairs in split_header_words([headervalue]):
            for name, value in pairs:
                cookie = requests.cookies.create_cookie(
                    name, value, domain=host, path=path
                )
                res.append(cookie)
    return res



# Copyright (C) 2022 Stefan fisk
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test srcset attribute parsing.
"""

from linkcheck.htmlutil.srcsetparse import parse_srcset

import pytest


# list of tuples
# (<input>, <expected parsed URLs>)
parsetests = [
    ('', []),
    ('   ', []),
    (',', []),
    ('\t\n ,,\t\n,,   \t\n', []),
    ('foo', ['foo']),
    ('foo,bar, ,foo, bar', ['foo,bar', 'foo', 'bar']),
    ('https://example.com/1 foo, https://example.com/2 bar',
        ['https://example.com/1', 'https://example.com/2']),
    ('   foo   ', ['foo']),
    (',,,foo,,,', ['foo']),
    (',foo,bar,baz,', ['foo,bar,baz']),
    ('foo bar baz', ['foo']),
    ('foo, bar baz', ['foo', 'bar']),
    ('foo/1 bar, foo/2', ['foo/1', 'foo/2']),
    ('foo/1 (foo/2)', ['foo/1']),
    ('foo/1 (((, foo/2', ['foo/1']),
]


class TestSrcsetParsing:
    @pytest.mark.parametrize("_in, _urls", parsetests)
    def test_parse(self, _in, _urls):
        assert parse_srcset(_in) == _urls



# Copyright (C) 2010-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test mime utility functions.
"""

import unittest
import os
from . import get_file
import linkcheck.mimeutil


class TestMiMeutil(unittest.TestCase):
    """Test file utility functions."""

    def mime_test(self, filename, mime_expected):
        absfilename = get_file(filename)
        with open(absfilename) as fd:
            mime = linkcheck.mimeutil.guess_mimetype(absfilename, read=fd.read)
        self.assertEqual(mime, mime_expected)

    def test_mime(self):
        filename = os.path.join("plist_binary", "Bookmarks.plist")
        self.mime_test(filename, "application/x-plist+safari")
        filename = os.path.join("plist_xml", "Bookmarks.plist")
        self.mime_test(filename, "application/x-plist+safari")
        self.mime_test("file.wml", "text/vnd.wap.wml")
        self.mime_test("sitemap.xml", "application/xml+sitemap")
        self.mime_test("sitemapindex.xml", "application/xml+sitemapindex")
        self.mime_test("no_decl_sitemap.xml", "application/xml+sitemap")
        self.mime_test("no_decl_sitemapindex.xml", "application/xml+sitemapindex")



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test robots.txt parsing.
"""

import unittest
from tests import need_network
from linkcheck import configuration, robotparser2

import requests


class TestRobotParser(unittest.TestCase):
    """
    Test robots.txt parser (needs internet access).
    """

    def setUp(self):
        """Initialize self.rp as a robots.txt parser."""
        self.rp = robotparser2.RobotFileParser(session=requests.Session())

    def check(self, a, b):
        """Helper function comparing two results a and b."""
        if not b:
            ac = "access denied"
        else:
            ac = "access allowed"
        if a != b:
            self.fail(f"{a} != {b} ({ac})")

    @need_network
    def test_nonexisting_robots(self):
        # robots.txt that does not exist
        self.rp.set_url("http://www.lycos.com/robots.txt")
        self.rp.read()
        self.check(
            self.rp.can_fetch(configuration.UserAgent, "http://www.lycos.com/search"),
            True,
        )

    @need_network
    def test_disallowed_robots(self):
        self.rp.set_url("http://google.com/robots.txt")
        self.rp.read()
        self.check(
            self.rp.can_fetch(configuration.UserAgent, "http://google.com/search"),
            False,
        )



# Copyright (C) 2005-2011 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test dummy object.
"""
# pylint: disable=no-member, not-callable

import unittest

import linkcheck.dummy


class TestDummy(unittest.TestCase):
    """
    Test dummy object.
    """

    def test_creation(self):
        dummy = linkcheck.dummy.Dummy()
        dummy = linkcheck.dummy.Dummy("1")
        dummy = linkcheck.dummy.Dummy("1", "2")
        dummy = linkcheck.dummy.Dummy(a=1, b=2)
        dummy = linkcheck.dummy.Dummy("1", a=None, b=2)
        del dummy

    def test_attributes(self):
        dummy = linkcheck.dummy.Dummy()
        dummy.hulla
        dummy.hulla.bulla
        dummy.hulla = 1
        del dummy.wulla
        del dummy.wulla.mulla

    def test_methods(self):
        dummy = linkcheck.dummy.Dummy()
        dummy.hulla()
        dummy.hulla().bulla()
        if "a" in dummy:
            pass

    def test_indexes(self):
        dummy = linkcheck.dummy.Dummy()
        len(dummy)
        dummy[1] = dummy[2]
        dummy[1][-1]
        dummy[1:3] = None
        del dummy[1]
        del dummy[2]
        del dummy[2:3]
        str(dummy)
        repr(dummy)



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Robots.txt parser.

The robots.txt Exclusion Protocol is implemented as specified in
https://www.robotstxt.org/norobots-rfc.txt
"""
import time
import urllib.parse

import requests

from . import log, LOG_CHECK, configuration

__all__ = ["RobotFileParser"]

ACCEPT_ENCODING = 'x-gzip,gzip,deflate'


class RobotFileParser:
    """This class provides a set of methods to read, parse and answer
    questions about a single robots.txt file."""

    def __init__(self, session, url='', auth=None, timeout=None):
        """Initialize internal entry lists and store given url and
        credentials."""
        self.set_url(url)
        self.session = session
        self.auth = auth
        self.timeout = timeout
        self._reset()

    def _reset(self):
        """Reset internal flags and entry lists."""
        self.entries = []
        self.default_entry = None
        self.disallow_all = False
        self.allow_all = False
        self.last_checked = 0
        # list of tuples (sitemap url, line number)
        self.sitemap_urls = []
        self.encoding = None

    def mtime(self):
        """Returns the time the robots.txt file was last fetched.

        This is useful for long-running web spiders that need to
        check for new robots.txt files periodically.

        @return: last modified in time.time() format
        @rtype: number
        """
        return self.last_checked

    def modified(self):
        """Set the time the robots.txt file was last fetched to the
        current time."""
        self.last_checked = time.time()

    def set_url(self, url):
        """Set the URL referring to a robots.txt file."""
        self.url = url
        self.host, self.path = urllib.parse.urlparse(url)[1:3]

    def read(self):
        """Read the robots.txt URL and feeds it to the parser."""
        self._reset()
        kwargs = dict(
            headers={
                'User-Agent': configuration.UserAgent,
                'Accept-Encoding': ACCEPT_ENCODING,
            }
        )
        if self.auth:
            kwargs["auth"] = self.auth
        if self.timeout:
            kwargs["timeout"] = self.timeout
        try:
            response = self.session.get(self.url, **kwargs)
            response.raise_for_status()
            log.debug(LOG_CHECK, "Robots response headers: %s", response.headers)
            content_type = response.headers.get('content-type')
            self.encoding = response.encoding = "utf-8"
            if content_type and content_type.lower().startswith('text/plain'):
                self.parse(response.iter_lines(decode_unicode=True))
            else:
                log.debug(LOG_CHECK, "%r allow all (no text content)", self.url)
                self.allow_all = True
        except requests.HTTPError as x:
            if x.response.status_code in (401, 403):
                self.disallow_all = True
                log.debug(
                    LOG_CHECK,
                    "%r disallow all (code %d)",
                    self.url,
                    x.response.status_code,
                )
            else:
                self.allow_all = True
                log.debug(LOG_CHECK, "%r allow all (HTTP error)", self.url)
        except requests.exceptions.Timeout:
            raise
        except requests.exceptions.RequestException:
            # no network or other failure
            self.allow_all = True
            log.debug(LOG_CHECK, "%r allow all (request error)", self.url)

    def _add_entry(self, entry):
        """Add a parsed entry to entry list.

        @return: None
        """
        if "*" in entry.useragents:
            # the default entry is considered last
            self.default_entry = entry
        else:
            self.entries.append(entry)

    def parse(self, lines):
        """Parse the input lines from a robot.txt file.
        We allow that a user-agent: line is not preceded by
        one or more blank lines.

        @return: None
        """
        log.debug(LOG_CHECK, "%r parse lines", self.url)
        state = 0
        linenumber = 0
        entry = Entry()

        for line in lines:
            line = line.strip()
            linenumber += 1
            if not line:
                if state == 1:
                    log.debug(
                        LOG_CHECK,
                        "%r line %d: allow or disallow directives without any"
                        " user-agent line",
                        self.url,
                        linenumber,
                    )
                    entry = Entry()
                    state = 0
                elif state == 2:
                    self._add_entry(entry)
                    entry = Entry()
                    state = 0
            # remove optional comment and strip line
            i = line.find('#')
            if i >= 0:
                line = line[:i]
            line = line.strip()
            if not line:
                continue
            line = line.split(':', 1)
            if len(line) == 2:
                line[0] = line[0].strip().lower()
                line[1] = urllib.parse.unquote(line[1].strip(), self.encoding)
                if line[0] == "user-agent":
                    if state == 2:
                        log.debug(
                            LOG_CHECK,
                            "%r line %d: missing blank line before"
                            " user-agent directive",
                            self.url,
                            linenumber,
                        )
                        self._add_entry(entry)
                        entry = Entry()
                    entry.useragents.append(line[1])
                    state = 1
                elif line[0] == "disallow":
                    if state == 0:
                        log.debug(
                            LOG_CHECK,
                            "%r line %d: missing user-agent directive before this line",
                            self.url,
                            linenumber,
                        )
                    else:
                        entry.rulelines.append(RuleLine(line[1], False))
                        state = 2
                elif line[0] == "allow":
                    if state == 0:
                        log.debug(
                            LOG_CHECK,
                            "%r line %d: missing user-agent directive before this line",
                            self.url,
                            linenumber,
                        )
                    else:
                        entry.rulelines.append(RuleLine(line[1], True))
                        state = 2
                elif line[0] == "crawl-delay":
                    if state == 0:
                        log.debug(
                            LOG_CHECK,
                            "%r line %d: missing user-agent directive before this line",
                            self.url,
                            linenumber,
                        )
                    else:
                        try:
                            entry.crawldelay = max(0, int(line[1]))
                            state = 2
                        except (ValueError, OverflowError):
                            log.debug(
                                LOG_CHECK,
                                "%r line %d: invalid delay number %r",
                                self.url,
                                linenumber,
                                line[1],
                            )
                elif line[0] == "sitemap":
                    # Note that sitemap URLs must be absolute according to
                    # http://www.sitemaps.org/protocol.html#submit_robots
                    # But this should be checked by the calling layer.
                    self.sitemap_urls.append((line[1], linenumber))
                else:
                    log.debug(
                        LOG_CHECK,
                        "%r line %d: unknown key %r",
                        self.url,
                        linenumber,
                        line[0],
                    )
            else:
                log.debug(
                    LOG_CHECK,
                    "%r line %d: malformed line %r",
                    self.url,
                    linenumber,
                    line,
                )
        if state in (1, 2):
            self.entries.append(entry)
        self.modified()
        log.debug(LOG_CHECK, "Parsed rules:\n%s", str(self))

    def can_fetch(self, useragent, url):
        """Using the parsed robots.txt decide if useragent can fetch url.

        @return: True if agent can fetch url, else False
        @rtype: bool
        """
        log.debug(
            LOG_CHECK,
            "%r check allowance for:\n  user agent: %r\n  url: %r ...",
            self.url,
            useragent,
            url,
        )
        if self.disallow_all:
            log.debug(LOG_CHECK, " ... disallow all.")
            return False
        if self.allow_all:
            log.debug(LOG_CHECK, " ... allow all.")
            return True
        # search for given user agent matches
        # the first match counts
        url = (
            urllib.parse.quote(urllib.parse.urlparse(urllib.parse.unquote(url))[2])
            or "/"
        )
        for entry in self.entries:
            if entry.applies_to(useragent):
                return entry.allowance(url)
        # try the default entry last
        if self.default_entry is not None:
            return self.default_entry.allowance(url)
        # agent not found ==> access granted
        log.debug(LOG_CHECK, " ... agent not found, allow.")
        return True

    def get_crawldelay(self, useragent):
        """Look for a configured crawl delay.

        @return: crawl delay in seconds or zero
        @rtype: integer >= 0
        """
        for entry in self.entries:
            if entry.applies_to(useragent):
                return entry.crawldelay
        return 0

    def __str__(self):
        """Constructs string representation, usable as contents of a
        robots.txt file.

        @return: robots.txt format
        @rtype: string
        """
        lines = [str(entry) for entry in self.entries]
        if self.default_entry is not None:
            lines.append(str(self.default_entry))
        return "\n\n".join(lines)


class RuleLine:
    """A rule line is a single "Allow:" (allowance==1) or "Disallow:"
    (allowance==0) followed by a path.
    """

    def __init__(self, path, allowance):
        """Initialize with given path and allowance info."""
        if path == '' and not allowance:
            # an empty value means allow all
            allowance = True
            path = '/'
        self.path = urllib.parse.quote(path)
        self.allowance = allowance

    def applies_to(self, path):
        """Look if given path applies to this rule.

        @return: True if pathname applies to this rule, else False
        @rtype: bool
        """
        return self.path == "*" or path.startswith(self.path)

    def __str__(self):
        """Construct string representation in robots.txt format.

        @return: robots.txt format
        @rtype: string
        """
        return ("Allow" if self.allowance else "Disallow") + ": " + self.path


class Entry:
    """An entry has one or more user-agents and zero or more rulelines."""

    def __init__(self):
        """Initialize user agent and rule list."""
        self.useragents = []
        self.rulelines = []
        self.crawldelay = 0

    def __str__(self):
        """string representation in robots.txt format.

        @return: robots.txt format
        @rtype: string
        """
        lines = ["User-agent: %s" % agent for agent in self.useragents]
        if self.crawldelay:
            lines.append("Crawl-delay: %d" % self.crawldelay)
        lines.extend([str(line) for line in self.rulelines])
        return "\n".join(lines)

    def applies_to(self, useragent):
        """Check if this entry applies to the specified agent.

        @return: True if this entry applies to the agent, else False.
        @rtype: bool
        """
        if not useragent:
            return True
        useragent = useragent.lower()
        for agent in self.useragents:
            if agent == '*':
                # we have the catch-all agent
                return True
            if agent.lower() in useragent:
                return True
        return False

    def allowance(self, filename):
        """Preconditions:
        - our agent applies to this entry
        - filename is URL decoded

        Check if given filename is allowed to access this entry.

        @return: True if allowed, else False
        @rtype: bool
        """
        for line in self.rulelines:
            log.debug(LOG_CHECK, "%s %s %s", filename, str(line), line.allowance)
            if line.applies_to(filename):
                log.debug(LOG_CHECK, " ... rule line %s", line)
                return line.allowance
        log.debug(
            LOG_CHECK,
            " ... no rule lines of %s applied to %s; allowed.",
            self.useragents,
            filename,
        )
        return True



# Copyright (C) 2004-2012 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test bookmark file parsing.
"""
from . import LinkCheckTest
from .. import need_network
import os


class TestBookmarks(LinkCheckTest):
    """
    Test bookmark link checking and content parsing.
    """

    @need_network
    def test_firefox_bookmarks(self):
        # firefox 3 bookmark file parsing
        self.file_test("places.sqlite")

    @need_network
    def test_opera_bookmarks(self):
        # Opera bookmark file parsing
        self.file_test("opera6.adr")

    @need_network
    def test_chromium_bookmarks(self):
        # Chromium and Google Chrome bookmark file parsing
        self.file_test("Bookmarks")

    @need_network
    def test_safari_bookmarks_xml(self):
        # Safari bookmark file parsing (for plaintext plist files)
        self.file_test(os.path.join("plist_xml", "Bookmarks.plist"))

    @need_network
    def test_safari_bookmarks_binary(self):
        # Safari bookmark file parsing (for binary plist files)
        self.file_test(os.path.join("plist_binary", "Bookmarks.plist"))



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Check page content for virus infection with clamav.
"""
import os
import socket
from . import _ContentPlugin
from .. import log, LOG_PLUGIN
from ..socketutil import create_socket


class VirusCheck(_ContentPlugin):
    """Checks the page content for virus infections with clamav.
    A local clamav daemon must be installed."""

    def __init__(self, config):
        """Initialize clamav configuration."""
        super().__init__(config)
        # XXX read config
        self.clamav_conf = get_clamav_conf(canonical_clamav_conf())
        if not self.clamav_conf:
            log.warn(LOG_PLUGIN, "clamav daemon not found for VirusCheck plugin")
        log.warn(
            LOG_PLUGIN, _("VirusCheck plugin does not support ClamAV >= 1.0.")
        )

    def applies_to(self, url_data):
        """Check for clamav and extern."""
        return False  # XXX Plugin disabled
        return self.clamav_conf and not url_data.extern[0]

    def check(self, url_data):
        """Scan content for viruses."""
        data = url_data.get_raw_content()
        infected, errors = scan(data, self.clamav_conf)
        if infected or errors:
            for msg in infected:
                url_data.add_warning("Virus scan infection: %s" % msg)
            for msg in errors:
                url_data.add_warning("Virus scan error: %s" % msg)
        else:
            url_data.add_info("No viruses in data found.")

    @classmethod
    def read_config(cls, configparser):
        """Read configuration file options."""
        config = dict()
        section = cls.__name__
        option = "clamavconf"
        if configparser.has_option(section, option):
            value = configparser.get(section, option)
        else:
            value = None
        config[option] = value
        return config


class ClamavError(Exception):
    """Raised on clamav errors."""

    pass


class ClamdScanner:
    """Virus scanner using a clamd daemon process."""

    def __init__(self, clamav_conf):
        """Initialize clamd daemon process sockets."""
        self.infected = []
        self.errors = []
        self.sock, self.host = clamav_conf.new_connection()
        self.sock_rcvbuf = self.sock.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)
        self.wsock = self.new_scansock()

    def new_scansock(self):
        """Return a connected socket for sending scan data to it."""
        port = None
        try:
            self.sock.sendall(b"STREAM")
            port = None
            for dummy in range(60):
                data = self.sock.recv(self.sock_rcvbuf)
                i = data.find(b"PORT")
                if i != -1:
                    port = int(data[i + 5:])
                    break
        except OSError:
            self.sock.close()
            raise
        if port is None:
            raise ClamavError(_("clamd is not ready for stream scanning"))
        sockinfo = get_sockinfo(self.host, port=port)
        wsock = create_socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            wsock.connect(sockinfo[0][4])
        except OSError:
            wsock.close()
            raise
        return wsock

    def scan(self, data):
        """Scan given data for viruses."""
        self.wsock.sendall(data)

    def close(self):
        """Get results and close clamd daemon sockets."""
        self.wsock.close()
        data = self.sock.recv(self.sock_rcvbuf)
        while data:
            if b"FOUND\n" in data:
                self.infected.append(data.decode('UTF-8', 'replace'))
            if b"ERROR\n" in data:
                self.errors.append(data.decode('UTF-8', 'replace'))
            data = self.sock.recv(self.sock_rcvbuf)
        self.sock.close()


def canonical_clamav_conf():
    """Default clamav configs for various platforms."""
    if os.name == 'posix':
        clamavconf = "/etc/clamav/clamd.conf"
    elif os.name == 'nt':
        clamavconf = r"c:\clamav-devel\etc\clamd.conf"
    else:
        clamavconf = "clamd.conf"
    return clamavconf


def get_clamav_conf(filename):
    """Initialize clamav configuration."""
    if os.path.isfile(filename):
        return ClamavConfig(filename)
    log.warn(LOG_PLUGIN, "No ClamAV config file found at %r.", filename)


def get_sockinfo(host, port=None):
    """Return socket.getaddrinfo for given host and port."""
    family, socktype = socket.AF_INET, socket.SOCK_STREAM
    return socket.getaddrinfo(host, port, family, socktype)


class ClamavConfig(dict):
    """Clamav configuration wrapper, with clamd connection method."""

    def __init__(self, filename):
        """Parse clamav configuration file."""
        super().__init__()
        self.parseconf(filename)
        if self.get('ScannerDaemonOutputFormat'):
            raise ClamavError(_("ScannerDaemonOutputFormat must be disabled"))
        if self.get('TCPSocket') and self.get('LocalSocket'):
            raise ClamavError(
                _("only one of TCPSocket and LocalSocket must be enabled")
            )

    def parseconf(self, filename):
        """Parse clamav configuration from given file."""
        with open(filename) as fd:
            # yet another config format, sigh
            for line in fd:
                line = line.strip()
                if not line or line.startswith("#"):
                    # ignore empty lines and comments
                    continue
                split = line.split(None, 1)
                if len(split) == 1:
                    self[split[0]] = True
                else:
                    self[split[0]] = split[1]

    def new_connection(self):
        """Connect to clamd for stream scanning.

        @return: tuple (connected socket, host)
        """
        if self.get('LocalSocket'):
            host = 'localhost'
            sock = self.create_local_socket()
        elif self.get('TCPSocket'):
            host = self.get('TCPAddr', 'localhost')
            sock = self.create_tcp_socket(host)
        else:
            raise ClamavError(_("one of TCPSocket or LocalSocket must be enabled"))
        return sock, host

    def create_local_socket(self):
        """Create local socket, connect to it and return socket object."""
        sock = create_socket(socket.AF_UNIX, socket.SOCK_STREAM)
        addr = self['LocalSocket']
        try:
            sock.connect(addr)
        except OSError:
            sock.close()
            raise
        return sock

    def create_tcp_socket(self, host):
        """Create tcp socket, connect to it and return socket object."""
        port = int(self['TCPSocket'])
        sockinfo = get_sockinfo(host, port=port)
        sock = create_socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect(sockinfo[0][4])
        except OSError:
            sock.close()
            raise
        return sock


def scan(data, clamconf):
    """Scan data for viruses.

    @return: (infection msgs, errors)
    @rtype: ([], [])
    """
    try:
        scanner = ClamdScanner(clamconf)
    except OSError:
        errmsg = _("Could not connect to ClamAV daemon.")
        return ([], [errmsg])
    try:
        scanner.scan(data)
    finally:
        scanner.close()
    return scanner.infected, scanner.errors



# Copyright (C) 2012-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Handle ignored URLs.
"""

from . import unknownurl


class IgnoreUrl(unknownurl.UnknownUrl):
    """Always ignored URL."""

    def is_ignored(self):
        """Return True if this URL scheme is ignored."""
        return True



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test filename routines.
"""

import unittest
import os
from linkcheck.checker.fileurl import get_nt_filename
from . import need_windows


class TestFilenames(unittest.TestCase):
    """
    Test filename routines.
    """

    @need_windows
    def test_nt_filename(self):
        path = os.getcwd()
        realpath = get_nt_filename(path)
        self.assertEqual(path, realpath)
        path = "c:\\"
        realpath = get_nt_filename(path)
        self.assertEqual(path, realpath)
        # XXX Only works on my computer.
        # Is there a Windows UNC share that is always available for tests?
        # path = '\\Vboxsrv\share\msg.txt'
        # realpath = get_nt_filename(path)
        # self.assertEqual(path, realpath)



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Base class for graph loggers.
"""
from . import _Logger
from ..decorators import notimplemented
import re


class _GraphLogger(_Logger):
    """Provide base method to get node data."""

    def __init__(self, **kwargs):
        """Initialize graph node list and internal id counter."""
        args = self.get_args(kwargs)
        super().__init__(**args)
        self.init_fileoutput(args)
        self.nodes = {}
        self.nodeid = 0

    def log_filter_url(self, url_data, do_print):
        """Update accounting data and log all valid URLs regardless the
        do_print flag.
        """
        self.stats.log_url(url_data, do_print)
        # ignore the do_print flag and determine ourselves if we filter the url
        if url_data.valid:
            self.log_url(url_data)

    def get_node(self, url_data):
        """Return new node data or None if node already exists."""
        if not url_data.url:
            return None
        elif url_data.url in self.nodes:
            return None
        node = {
            "url": url_data.url,
            "parent_url": url_data.parent_url,
            "id": self.nodeid,
            "label": quote(url_data.title if url_data.title else url_data.name),
            "extern": 1 if url_data.extern else 0,
            "checktime": url_data.checktime,
            "size": url_data.size,
            "dltime": url_data.dltime,
            "edge": quote(url_data.name),
            "valid": 1 if url_data.valid else 0,
        }
        self.nodes[node["url"]] = node
        self.nodeid += 1
        return node

    def write_edges(self):
        """
        Write all edges we can find in the graph in a brute-force manner.
        """
        for node in self.nodes.values():
            if node["parent_url"] in self.nodes:
                self.write_edge(node)
        self.flush()

    @notimplemented
    def write_edge(self, node):
        """Write edge data for one node and its parent."""
        pass

    @notimplemented
    def end_graph(self):
        """Write end-of-graph marker."""
        pass

    def end_output(self, **kwargs):
        """Write edges and end of checking info as gml comment."""
        self.write_edges()
        self.end_graph()
        if self.has_part("outro"):
            self.write_outro()
        self.close_fileoutput()


_disallowed = re.compile(r"[^a-zA-Z0-9 '#(){}\-\[\]\.,;:\!\?]+")


def quote(s):
    """Replace disallowed characters in node or edge labels.
    Also remove whitespace from beginning or end of label."""
    return _disallowed.sub(" ", s).strip()



# Copyright (C) 2017 Petr Dlouhý
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import unittest
from collections import namedtuple

import linkcheck.configuration
from linkcheck.cache.results import ResultCache
from linkcheck.cache.urlqueue import Empty, NUM_PUTS_CLEANUP, UrlQueue

UrlData = namedtuple("UrlData", "url cache_url aggregate has_result")
Aggregate = namedtuple("Aggregate", "result_cache")


class TestUrlQueue(unittest.TestCase):
    def setUp(self):
        config = linkcheck.configuration.Configuration()
        self.result_cache = ResultCache(config["resultcachesize"])
        self.urlqueue = UrlQueue()
        self.urldata1 = UrlData(
            url="Foo",
            cache_url="Foo",
            aggregate=Aggregate(result_cache=self.result_cache),
            has_result=True,
        )

    def test_max_allowed_urls_bad_value(self):
        with self.assertRaises(ValueError):
            UrlQueue(max_allowed_urls=0)
        with self.assertRaises(ValueError):
            UrlQueue(max_allowed_urls=-1)

    def test_qsize(self):
        """ Test qsize() """
        self.assertEqual(self.urlqueue.qsize(), 0)
        self.urlqueue.put(self.urldata1)
        self.assertEqual(self.urlqueue.qsize(), 1)

    def test_empty(self):
        """ Test empty() """
        self.assertEqual(self.urlqueue.empty(), True)
        self.urlqueue.put(self.urldata1)
        self.assertEqual(self.urlqueue.empty(), False)

    def test_get_empty(self):
        """ Test, that get() with empty queue throws Empty """
        with self.assertRaises(Empty):
            self.assertEqual(self.urlqueue.get(0), None)

    def test_get_negative_timeout(self):
        """ Test, that get() with negative timeout throws ValueError """
        with self.assertRaises(ValueError):
            self.assertEqual(self.urlqueue.get(-1), None)

    def test_put_get(self):
        """
        Test, that after put() we can get()
        the item and it can be get only once
        """
        self.urlqueue.put(self.urldata1)
        cached_item = self.result_cache.get_result(self.urldata1)
        self.assertEqual(cached_item, None)
        self.assertEqual(self.urlqueue.get(), self.urldata1)
        with self.assertRaises(Empty):
            self.assertEqual(self.urlqueue.get(0), None)

    def test_put_has_result_false(self):
        """
        Test, that element with has_result=False
        is put() on the end of queue
        """
        self.urlqueue.put(self.urldata1)
        urldata = UrlData(
            url="Bar",
            cache_url="Bar",
            aggregate=Aggregate(result_cache=self.result_cache),
            has_result=False,
        )
        self.urlqueue.put(urldata)
        self.assertEqual(self.urlqueue.get(), self.urldata1)
        self.assertEqual(self.urlqueue.get(), urldata)
        with self.assertRaises(Empty):
            self.assertEqual(self.urlqueue.get(0), None)

    def test_put_has_result_true(self):
        """
        Test, that element with has_result=True
        is put() on the beginning of queue
        """
        self.urlqueue.put(self.urldata1)
        urldata = UrlData(
            url="Bar",
            cache_url="Bar",
            aggregate=Aggregate(result_cache=self.result_cache),
            has_result=True,
        )
        self.urlqueue.put(urldata)
        self.assertEqual(self.urlqueue.get(), urldata)
        self.assertEqual(self.urlqueue.get(), self.urldata1)
        with self.assertRaises(Empty):
            self.assertEqual(self.urlqueue.get(0), None)

    def test_put_cache(self):
        """
        Test, that making put() on two elements with same
        cache_url adds only one element
        """
        self.urlqueue.put(self.urldata1)
        urldata = UrlData(
            url="Bar",
            cache_url="Foo",
            aggregate=Aggregate(result_cache=self.result_cache),
            has_result=True,
        )
        self.urlqueue.put(urldata)
        self.assertEqual(self.urlqueue.qsize(), 1)
        self.assertEqual(self.urlqueue.get(), self.urldata1)
        with self.assertRaises(Empty):
            self.assertEqual(self.urlqueue.get(0), None)

    def test_cleanup(self):
        """
        Test, that after adding NUM_PUTS_CLEANUP elements
        the queue is cleaned up.
        Whether the cleanup is was performed is determined,
        that element in cache is now on top of the queue.
        """
        for i in range(NUM_PUTS_CLEANUP - 1):
            self.urlqueue.put(
                UrlData(
                    url="Bar",
                    cache_url="Bar address %s" % i,
                    aggregate=Aggregate(result_cache=self.result_cache),
                    has_result=False,
                ),
            )
        self.assertEqual(self.urlqueue.qsize(), NUM_PUTS_CLEANUP - 1)
        urldata = UrlData(
            url="Bar",
            cache_url="Bar address",
            aggregate=Aggregate(result_cache=self.result_cache),
            has_result=False,
        )
        self.result_cache.add_result("Bar address 2", "asdf")
        self.urlqueue.put(urldata)
        self.assertEqual(self.urlqueue.qsize(), NUM_PUTS_CLEANUP)
        self.assertEqual(self.urlqueue.get().cache_url, "Bar address 2")



#!/usr/bin/env python
# Copyright (C) 2012-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Analyze a memory dump by the meliae module.
"""
import sys
import os
import codecs
import html
from linkcheck import strformat


def main(filename):
    om = print_memorydump(filename)
    dirname, basename = os.path.split(filename)
    basename = os.path.splitext(basename)[0]
    basedir = os.path.join(dirname, basename)
    if not os.path.isdir(basedir):
        os.mkdir(basedir)
    write_htmlfiles(om, basedir)


def print_memorydump(filename):
    from meliae import loader

    om = loader.load(filename, collapse=True)
    om.remove_expensive_references()
    print(om.summarize())
    return om


def write_htmlfiles(om, basedir):
    om.compute_parents()
    open_files = {}
    for obj in om.objs.itervalues():
        fp = get_file(obj.type_str, open_files, basedir)
        write_html_obj(fp, obj, om.objs)
    close_files(open_files)


def get_file(type_str, open_files, basedir):
    """Get already opened file, or open and initialize a new one."""
    if type_str not in open_files:
        filename = type_str + ".html"
        encoding = "utf-8"
        fd = codecs.open(os.path.join(basedir, filename), "w", encoding)
        open_files[type_str] = fd
        write_html_header(fd, type_str, encoding)
    return open_files[type_str]


def close_files(open_files):
    for fp in open_files.values():
        write_html_footer(fp)
        fp.close()


HtmlHeader = """
<!doctype html>
<head>
    <meta charset="%s">
</head>
<body>
"""


def write_html_header(fp, type_str, encoding):
    fp.write(HtmlHeader % encoding)
    fp.write("<h1>Type %s</h1>\n" % type_str)
    fp.write(
        "<table><tr><th>Address</th><th>Name</th><th>Size</th><th>Parents</th>"
        "<th>References</th></tr>\n"
    )


def get_children(obj, objs):
    res = []
    for address in obj.children:
        if address in objs:
            child = objs[address]
            url = "#%d" % address
            if child.type_str != obj.type_str:
                url = child.type_str + ".html" + url
            entry = '<a href="%s">%d</a>' % (url, address)
        else:
            entry = "%d" % address
        res.append(entry)
    return res


def get_parents(obj, objs):
    res = []
    for address in obj.parents:
        if address in objs:
            parent = objs[address]
            url = "#%d" % address
            if parent.type_str != obj.type_str:
                url = parent.type_str + ".html" + url
            entry = '<a href="%s">%d</a>' % (url, address)
        else:
            entry = "%d" % address
        res.append(entry)
    return res


def write_html_obj(fp, obj, objs):
    if obj.value is None:
        value = "None"
    else:
        value = html.escape(str(obj.value))
    attrs = dict(
        address=obj.address,
        size=strformat.strsize(obj.size),
        children=",".join(get_children(obj, objs)),
        parents=",".join(get_parents(obj, objs)),
        value=value,
    )
    fp.write(
        "<tr><td>%(address)d</td><td>%(value)s</td><td>%(size)s</td>"
        "<td>%(children)s</td></tr>\n" % attrs
    )


def write_html_footer(fp):
    fp.write("</table></body></html>")


if __name__ == "__main__":
    filename = sys.argv[1]
    main(filename)



# -*- coding: iso-8859-1 -*-
# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test error checking.
"""
from . import LinkCheckTest


class TestError(LinkCheckTest):
    """
    Test unrecognized or syntactically wrong links.
    """

    def test_unrecognized(self):
        # Unrecognized scheme
        url = "hutzli:"
        attrs = self.get_attrs(url=url)
        attrs["nurl"] = self.norm("file://%(curdir)s/%(url)s" % attrs)
        resultlines = [
            "url file://%(curdir)s/%(url)s" % attrs,
            "cache key %(nurl)s" % attrs,
            "real url %(nurl)s" % attrs,
            "error",
        ]
        self.direct(url, resultlines)

    def test_invalid1(self):
        # invalid scheme chars
        url = "äöü:"
        attrs = self.get_attrs(url=url)
        attrs["nurl"] = self.norm("file://%(curdir)s/%(url)s" % attrs)
        resultlines = [
            "url file://%(curdir)s/%(url)s" % attrs,
            "cache key %(nurl)s" % attrs,
            "real url %(nurl)s" % attrs,
            "name %(url)s" % attrs,
            "error",
        ]
        self.direct(url, resultlines)

    def test_invalid2(self):
        # missing scheme alltogether
        url = "äöü"
        attrs = self.get_attrs(url=url)
        attrs["nurl"] = self.norm("file://%(curdir)s/%(url)s" % attrs)
        resultlines = [
            "url file://%(curdir)s/%(url)s" % attrs,
            "cache key %(nurl)s" % attrs,
            "real url %(nurl)s" % attrs,
            "name %(url)s" % attrs,
            "error",
        ]
        self.direct(url, resultlines)

    def test_invalid3(self):
        # really fucked up
        url = "@³²¼][½ ³@] ¬½"
        attrs = self.get_attrs(url=url)
        attrs["nurl"] = self.norm("file://%(curdir)s/%(url)s" % attrs)
        resultlines = [
            "url file://%(curdir)s/%(url)s" % attrs,
            "cache key %(nurl)s" % attrs,
            "real url %(nurl)s" % attrs,
            "name %(url)s" % attrs,
            "error",
        ]
        self.direct(url, resultlines)



# Copyright (C) 2000-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Application internationalization support.
"""

# i18n support
import locale
import gettext
import sys
import codecs

default_encoding = None


def init(domain, directory):
    """Initialize this gettext i18n module and install the gettext translator class."""
    global default_encoding
    default_encoding = locale.getpreferredencoding(False)

    translator = gettext.translation(domain, localedir=directory, fallback=True)

    import builtins
    builtins.__dict__['_'] = translator.gettext
    builtins.__dict__['_n'] = translator.ngettext


def get_encoded_writer(out=sys.stdout, encoding=None, errors='replace'):
    """Get wrapped output writer with given encoding and error handling."""
    if encoding is None:
        encoding = default_encoding
    Writer = codecs.getwriter(encoding)
    return Writer(out.buffer, errors)



# Copyright (C) 2010-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test checking of unknown URLs.
"""
from . import LinkCheckTest


class TestUnknown(LinkCheckTest):
    """Test unknown URL scheme checking."""

    def test_skype(self):
        url = "skype:"
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "info Skype URL ignored.",
            "valid",
        ]
        self.direct(url, resultlines)

    def test_irc(self):
        url = "irc://example.org"
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "info Irc URL ignored.",
            "valid",
        ]
        self.direct(url, resultlines)
        url = "ircs://example.org"
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "info Ircs URL ignored.",
            "valid",
        ]
        self.direct(url, resultlines)

    def test_steam(self):
        url = "steam://connect/example.org"
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "info Steam URL ignored.",
            "valid",
        ]
        self.direct(url, resultlines)

    def test_feed(self):
        url = "feed:https://example.com/entries.atom"
        nurl = "feed:https%3A/example.com/entries.atom"
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "info Feed URL ignored.",
            "valid",
        ]
        self.direct(url, resultlines)
        url = "feed://example.com/entries.atom"
        nurl = self.norm(url)
        resultlines = [
            "url %s" % url,
            "cache key %s" % nurl,
            "real url %s" % nurl,
            "info Feed URL ignored.",
            "valid",
        ]
        self.direct(url, resultlines)



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test whitespace handling.
"""
from tests import need_network
from . import LinkCheckTest


class TestWhitespace(LinkCheckTest):
    """
    Test whitespace in URLs.
    """

    @need_network
    def test_leading_whitespace(self):
        # Leading whitespace
        url = " https://linkchecker.github.io/"
        attrs = self.get_attrs(url=url)
        attrs["surl"] = url.strip()
        resultlines = [
            "url %(surl)s" % attrs,
            "cache key %(surl)s" % attrs,
            "real url %(surl)s" % attrs,
            "warning Leading or trailing whitespace in URL `%(url)s'." % attrs,
            "valid",
        ]
        self.direct(url, resultlines)
        url = "\nhttps://linkchecker.github.io/"
        attrs = self.get_attrs(url=url)
        attrs["surl"] = url.strip()
        resultlines = [
            "url %(surl)s" % attrs,
            "cache key %(surl)s" % attrs,
            "real url %(surl)s" % attrs,
            "warning Leading or trailing whitespace in URL `%(url)s'." % attrs,
            "valid",
        ]
        self.direct(url, resultlines)

    @need_network
    def test_trailing_whitespace(self):
        # Trailing whitespace
        url = "https://linkchecker.github.io/ "
        resultlines = [
            "url %s" % url.strip(),
            "cache key %s" % url.strip(),
            "real url %s" % url.strip(),
            "warning Leading or trailing whitespace in URL `%s'." % url,
            "valid",
        ]
        self.direct(url, resultlines)
        url = "https://linkchecker.github.io/\n"
        resultlines = [
            "url %s" % url.strip(),
            "cache key %s" % url.strip(),
            "real url %s" % url.strip(),
            "warning Leading or trailing whitespace in URL `%s'." % url,
            "valid",
        ]
        self.direct(url, resultlines)



# Copyright (C) 2004-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test http checking.
"""
import sys

from tests import need_network
from .httpserver import HttpServerTest, CookieRedirectHttpRequestHandler


class TestHttp(HttpServerTest):
    """Test http:// link checking."""

    def __init__(self, methodName="runTest"):
        super().__init__(methodName=methodName)
        self.handler = CookieRedirectHttpRequestHandler

    @need_network
    def test_html_internet(self):
        confargs = dict(recursionlevel=1)
        self.file_test("http.html", confargs=confargs)
        self.file_test("http_lowercase.html", confargs=confargs)
        self.file_test("http_quotes.html", confargs=confargs)
        if sys.version_info < (3, 12, 6):
            http_slash_html = "http_slash-old-urlunsplit.html"
        else:
            http_slash_html = "http_slash.html"
        self.file_test(http_slash_html, confargs=confargs)
        self.file_test("http_url_quote.html", confargs=confargs)

    def test_html(self):
        confargs = dict(recursionlevel=1)
        self.file_test("http_empty.html", confargs=confargs)
        self.file_test("http_file.html", confargs=confargs)
        self.file_test("http_utf8.html", confargs=confargs)
        self.file_test("http.xhtml", confargs=confargs)
        self.file_test("http_invalid_host.html", confargs=confargs)

    def test_status(self):
        for status in sorted(self.handler.responses.keys()):
            self._test_status(status)

    def _test_status(self, status):
        url = "http://localhost:%d/status/%d" % (self.port, status)
        resultlines = [
            "url %s" % url,
            "cache key %s" % url,
            "real url %s" % url,
        ]
        if status in (204,):
            resultlines.append("warning No Content")
        if status == 429:
            resultlines.append("warning Rate limited (Retry-After: None)")
        if (status not in [101, 102, 103] and status < 200) or (
            status >= 400 and status != 429
        ):
            result = "error"
        else:
            result = "valid"
        resultlines.append(result)
        self.direct(url, resultlines, recursionlevel=0)



# Copyright (C) 2007-2009 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test internal pattern construction
"""
import linkcheck.director
import linkcheck.configuration
from .__init__ import LinkCheckTest, get_url_from


class TestInternpat(LinkCheckTest):
    """Test internal pattern."""

    def test_trailing_slash(self):
        # Make sure a trailing slash is not lost
        config = linkcheck.configuration.Configuration()
        aggregate = linkcheck.director.get_aggregate(config)
        url = "http://example.org/foo/"
        url_data = get_url_from(url, 0, aggregate)
        internpat = url_data.get_intern_pattern()
        self.assertTrue(internpat.endswith("/"))



# Copyright (C) 2006-2014 Bastian Kleineidam
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
Test virus filter.
"""
import unittest
from tests import need_clamav
from linkcheck.plugins import viruscheck as clamav


class TestClamav(unittest.TestCase):
    def setUp(self):
        self.clamav_conf = clamav.get_clamav_conf("/etc/clamav/clamd.conf")

    @need_clamav
    def testClean(self):
        data = b""
        infected, errors = clamav.scan(data, self.clamav_conf)
        self.assertFalse(infected)
        self.assertFalse(errors)

    @need_clamav
    def testInfected(self):
        # from the clamav test directory: the clamav test file as html data
        data = (
            b'<a href="data:application/octet-stream;base64,'
            b"TVpQAAIAAAAEAA8A//8AALgAAAAhAAAAQAAaAAAAAAAAAAAAAAAAAAAAAAAAAA"
            b"AAAAAAAAAAAAAAAAAAAAEAALtxEEAAM8BQUIvzU1NQsClAMARmrHn5ujEAeA2t"
            b"UP9mcA4fvjEA6eX/tAnNIbRMzSFiDAoBAnB2FwIeTgwEL9rMEAAAAAAAAAAAAA"
            b"AAAAAAwBAAAIAQAAAAAAAAAAAAAAAAAADaEAAA9BAAAAAAAAAAAAAAAAAAAAAA"
            b"AAAAAAAAS0VSTkVMMzIuRExMAABFeGl0UHJvY2VzcwBVU0VSMzIuRExMAENMQU"
            b"1lc3NhZ2VCb3hBAOYQAAAAAAAAPz8/P1BFAABMAQEAYUNhQgAAAAAAAAAA4ACO"
            b"gQsBAhkABAAAAAYAAAAAAABAEAAAABAAAEAAAAAAAEAAABAAAAACAAABAAAAAA"
            b"AAAAMACgAAAAAAACAAAAAEAAAAAAAAAgAAAAAAEAAAIAAAAAAQAAAQAAAAAAAA"
            b"EAAAAAAAAAAAAAAAhBAAAIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
            b"AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
            b"AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAW0NMQU1BVl"
            b'0AEAAAABAAAAACAAABAAAAAAAAAAAAAAAAAAAAAAAAwA==">t</a>'
        )
        infected, errors = clamav.scan(data, self.clamav_conf)
        # different versions of clamav report different responses, apparently
        acceptable_responses = (
            [
              "stream: ClamAV-Test-File(2d1206194bd704385e37000be6113f73:781) FOUND\n"
            ],
            [
              "stream: Clamav.Test.File-6(aa15bcf478d165efd2065190eb473bcb:544) FOUND\n"
            ],
            [
              "stream: Clamav.Test.File-6(2d1206194bd704385e37000be6113f73:781) FOUND\n"
            ],
        )
        self.assertIn(infected, acceptable_responses)
        self.assertFalse(errors)
